/******************************************************************************
 * Copyright (c) 2016 Oracle
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/

package oracle.eclipse.tools.common.upgrade;

import java.util.Date;
import java.util.List;

import org.eclipse.core.resources.IProject;

/**
 * Project upgrade log contains the record of which upgrade participants have executed on the
 * project during it's existence. The log is kept in project metadata. 
 * 
 * <p>This interface should not be implemented outside of the upgrade framework.</p> 
 *  
 * @author <a href="mailto:konstantin.komissarchik@oracle.com">Konstantin Komissarchik</a>
 */

public interface IProjectUpgradeLog
{
    /**
     * Represents a single entry in the log. 
     */
    
    interface ILogEntry
    {
        /**
         * Returns the time of recording of this log entry.
         *  
         * @return the time of recording of this log entry
         */
        
        Date getTime();
        
        /**
         * Returns the information about the upgrade participant that was executed.
         * 
         * @return the information about the upgrade participant that was executed
         */
        
        IParticipantRecord getParticipantRecord();
    }
    
    /**
     * Describes an upgrade participant. Each log entry contains a list of 
     * <code>IParticipantRecord</code> objects to represent upgrade participants that were 
     * executed.
     */
    
    interface IParticipantRecord
    {
        /**
         * Returns the id of the plugin that the upgrade participant was contributed from.
         *
         * @return the id of the plugin that the upgrade participant was contributed from
         */
        
        String getPluginId();
        
        /**
         * Returns the version of the plugin that the upgrade participant was contributed from.
         * 
         * @return the version of the plugin that the upgrade participant was contributed from
         */
        
        String getPluginVersion();
        
        /**
         * Returns the class name of the upgrade participant.
         * 
         * @return the class name of the upgrade participant
         */
        
        String getClassName();
    }
    
    /**
     * Returns the project that this upgrade log belongs to.
     * 
     * @return the project that this upgrade log belongs to
     */
    
    IProject getProject();
    
    /**
     * Returns the list of entries contained in this log. The entries are arranged in a 
     * chronological order.
     * 
     * @return a read-only list of entries contained in this log
     */
    
    List<ILogEntry> getEntries();

    /**
     * Finds the last log entry that references the specified upgrade participant class name.
     * 
     * @param className the upgrade participant class name to look for in the log
     * @return the last log entry that references specified upgrade participant class name or
     *   <code>null</code> if not found
     */
    
    ILogEntry findLastLogEntry( String className );
    
}
