/******************************************************************************
 * Copyright (c) 2016 Oracle
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/

package oracle.eclipse.tools.common.upgrade;

import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;

/**
 * The interface implemented in order to participate in project upgrade.
 * 
 * @author <a href="mailto:konstantin.komissarchik@oracle.com">Konstantin Komissarchik</a>
 */

public interface IProjectUpgradeParticipant 
{
    public class Action
    {
        public enum Type
        {
            CREATE,
            DELETE,
            EDIT,
            CHANGE_MARKERS
        }
        
        private final Type type;
        private final IFile file;
        
        public Action( final Type type )
        {
        	if( type == null )
        	{
        		throw new NullPointerException();
        	}
            
        	this.type = type;
        	this.file = null;
        }
        
        public Action( final Type type,
                       final IFile file )
        {
            if( type == null || file == null )
            {
                throw new NullPointerException();
            }
            
            this.type = type;
            this.file = file;
        }
        
        public Type getType()
        {
            return this.type;
        }
        
        public IFile getFile()
        {
            return this.file;
        }
        
        @Override
        public boolean equals( final Object obj )
        {
            if( ! ( obj instanceof Action ) )
            {
                return false;
            }
            else
            {
                final Action action = (Action) obj;
                
                return this.type == action.type &&
                       this.file.equals( action.file );
            }
        }
        
        @Override
        public int hashCode()
        {
        	if (file == null)
        		return this.type.hashCode();
        	else
                return this.type.hashCode() ^ this.file.hashCode();
        }

        @Override
        public String toString()
        {
            final StringBuilder buf = new StringBuilder();
            
            buf.append( this.type.name() );
            buf.append( " [" ); //$NON-NLS-1$
            buf.append( this.file.getLocation().toOSString() );
            buf.append( ']' );
            
            return buf.toString();
        }
    }
    
    /**
     * This method is invoked by the upgrade framework to determine whether the
     * upgrade participant has any work to do. The implementation should be as
     * fast as possible since this method will be frequently invoked.
     * 
     * @param project the project
     * @return <code>true</code> if the upgrade participant has detected that
     *   some upgrade actions need to be taken on the project
     * @throws CoreException if failed for any reason
     */
    
    boolean check( IProject project )
        
        throws CoreException;
    
    Set<Action> prepare( IProject project )
    
        throws CoreException;
    
    /**
     * This method is invoked by the upgrade framework to apply upgrade actions
     * to the project. 
     * 
     * @param project the project
     * @param monitor the progress monitor, or <code>null</code> if progress
     *   reporting and ability to cancel is not necessary
     * @throws CoreException if failed for any reason
     */
    
    void execute( IProject project,
                  IProgressMonitor monitor )
    
        throws CoreException;
    
}
