/******************************************************************************
 * Copyright (c) 2016 Oracle
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/

package oracle.eclipse.tools.common.upgrade.internal;

import static oracle.eclipse.tools.util.DomUtil.element;
import static oracle.eclipse.tools.util.DomUtil.elements;
import static oracle.eclipse.tools.util.DomUtil.root;
import static oracle.eclipse.tools.util.EclipseFileUtil.writeFile;
import static oracle.eclipse.tools.util.MiscUtil.checkNotNull;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringWriter;
import java.nio.charset.StandardCharsets;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;
import java.util.concurrent.CopyOnWriteArrayList;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.w3c.dom.Element;

import oracle.eclipse.tools.common.upgrade.IProjectUpgradeLog;
import oracle.eclipse.tools.util.XmlParseException;
import oracle.eclipse.tools.util.XmlWriter;

/**
 * @author <a href="mailto:konstantin.komissarchik@oracle.com">Konstantin Komissarchik</a>
 */

public final class ProjectUpgradeLog implements IProjectUpgradeLog
{
    private static final String FILE_NAME 
        = ".settings/oracle.eclipse.tools.common.upgrade.log.xml"; //$NON-NLS-1$
    
    private static final String EL_ENTRY = "entry"; //$NON-NLS-1$
    private static final String EL_UPGRADE_LOG = "upgrade-log"; //$NON-NLS-1$
    private static final String EL_UPGRADE_PARTICIPANT = "upgrade-participant"; //$NON-NLS-1$
    private static final String ATTR_CLASS = "class"; //$NON-NLS-1$
    private static final String ATTR_PLUGIN_ID = "pluginId"; //$NON-NLS-1$
    private static final String ATTR_PLUGIN_VERSION = "pluginVersion"; //$NON-NLS-1$
    private static final String ATTR_TIME = "time"; //$NON-NLS-1$
    
    private static final class LogEntry
    
        implements ILogEntry
        
    {
        private static final DateFormat TIME_FORMAT;
        
        static
        {
            TIME_FORMAT = new SimpleDateFormat( "yyyy-MM-dd HH:mm:ss z" ); //$NON-NLS-1$
            TIME_FORMAT.setTimeZone( TimeZone.getTimeZone( "UTC" ) ); //$NON-NLS-1$
        }

        private final Date time;
        private final ParticipantRecord participantRecord;
        
        public LogEntry( final ParticipantRecord participantRecord )
        {
            this( new Date(), participantRecord );
        }
        
        public LogEntry( final Date time,
                         final ParticipantRecord participantRecord )
        {
            this.time = checkNotNull( time );
            this.participantRecord = checkNotNull( participantRecord );
        }
        
        public Date getTime()
        {
            return this.time;
        }
        
        public IParticipantRecord getParticipantRecord()
        {
            return this.participantRecord;
        }
        
        public void write( final XmlWriter out )
        
            throws IOException
            
        {
            out.startElement( EL_ENTRY );
            out.addAttribute( ATTR_TIME, convert( this.time ) );
            this.participantRecord.write( out );
            out.endElement();
        }
        
        public static LogEntry read( final Element element )
        {
            final Date time = convert( element.getAttribute( ATTR_TIME ).trim() );
            final Element elParticipantRecord = element( element, EL_UPGRADE_PARTICIPANT );
            
            if( time != null && elParticipantRecord != null )
            {
                final ParticipantRecord precord = ParticipantRecord.read( elParticipantRecord );
                
                if( precord != null )
                {
                    return new LogEntry( time, precord );
                }
            }
            
            return null;
        }
        
        private static String convert( final Date time )
        {
            return TIME_FORMAT.format( time );
        }
        
        private static Date convert( final String time )
        {
            try
            {
                return TIME_FORMAT.parse( time );
            }
            catch( ParseException e )
            {
                return null;
            }
        }
    }
    
    private static final class ParticipantRecord
    
        implements IParticipantRecord
        
    {
        private final String pluginId;
        private final String pluginVersion;
        private final String className;
        
        public ParticipantRecord( final String pluginId,
                                  final String pluginVersion,
                                  final String className )
        {
            this.pluginId = checkNotNull( pluginId );
            this.pluginVersion = checkNotNull( pluginVersion );
            this.className = checkNotNull( className );
        }
        
        public String getClassName()
        {
            return this.className;
        }

        public String getPluginId()
        {
            return this.pluginId;
        }

        public String getPluginVersion()
        {
            return this.pluginVersion;
        }
        
        public void write( final XmlWriter out )
            
            throws IOException
            
        {
            out.startElement( EL_UPGRADE_PARTICIPANT );
            out.addAttribute( ATTR_PLUGIN_ID, this.pluginId );
            out.addAttribute( ATTR_PLUGIN_VERSION, this.pluginVersion );
            out.addAttribute( ATTR_CLASS, this.className );
            out.endElement();
        }
        
        public static ParticipantRecord read( final Element element )
        {
            final String pluginId = element.getAttribute( ATTR_PLUGIN_ID ).trim();
            final String pluginVersion = element.getAttribute( ATTR_PLUGIN_VERSION ).trim();
            final String className = element.getAttribute( ATTR_CLASS ).trim();
            
            if( pluginId.length() == 0 || pluginVersion.length() == 0 || className.length() == 0 )
            {
                return null;
            }
            
            return new ParticipantRecord( pluginId, pluginVersion, className );
        }
    }
    
    private final IProject project;
    private final IFile logFile;
    private long logFileModificationStamp;
    private final List<ILogEntry> entries;
    
    public ProjectUpgradeLog( final IProject project )
    {
        this.project = checkNotNull( project );
        this.logFile = this.project.getFile( FILE_NAME );
        this.logFileModificationStamp = IResource.NULL_STAMP;
        this.entries = new CopyOnWriteArrayList<ILogEntry>();
        
        refresh();
    }
    
    public IProject getProject()
    {
        return this.project;
    }
    
    public IFile getLogFile()
    {
        return this.logFile;
    }
    
    public List<ILogEntry> getEntries()
    {
        return Collections.unmodifiableList( this.entries );
    }
    
    public ILogEntry findLastLogEntry( final String className )
    {
        synchronized( this )
        {
            for( int i = this.entries.size() - 1; i >= 0; i-- )
            {
                final ILogEntry entry = this.entries.get( i );
                
                if( entry.getParticipantRecord().getClassName().equals( className ) )
                {
                    return entry;
                }
            }
            
            return null;
        }
    }
    
    public void addEntry( final String pluginId,
                          final String pluginVersion,
                          final String className )
    {
        final ILogEntry entry 
            = new LogEntry( new ParticipantRecord( pluginId, pluginVersion, className ) );
        
        this.entries.add( entry );
        write();
    }

    public void refresh()
    {
        synchronized( this )
        {
            final long newModificationStamp = this.logFile.getModificationStamp();
            
            if( this.logFileModificationStamp == newModificationStamp )
            {
                return;
            }
            
            final List<ILogEntry> newLogEntries = new ArrayList<ILogEntry>();
            
            if( this.logFile.isAccessible() )
            {
                try( Reader reader = new InputStreamReader( this.logFile.getContents(), StandardCharsets.UTF_8 ) )
                {
                    for( Element elEntry : elements( root( reader ), EL_ENTRY ) )
                    {
                        final LogEntry entry = LogEntry.read( elEntry );
                        
                        if( entry != null )
                        {
                            newLogEntries.add( entry );
                        }
                        else
                        {
                            // todo: report error
                        }
                    }
                }
                catch( CoreException e )
                {
                    UpgradePlugin.log( e );
                }
                catch( XmlParseException e )
                {
                    UpgradePlugin.log( e );
                }
                catch( IOException e )
                {
                    UpgradePlugin.log( e );
				}
            }
            
            this.entries.clear();
            this.entries.addAll( newLogEntries );
            this.logFileModificationStamp = newModificationStamp;
        }
    }

    private void write()
    {
        final StringWriter sw = new StringWriter();
        
        try
        {
            final XmlWriter xmlw = new XmlWriter( sw );
            
            xmlw.startElement( EL_UPGRADE_LOG );
            
            for( ILogEntry entry : this.entries )
            {
                ( (LogEntry) entry ).write( xmlw );
            }
            
            xmlw.endElement();
        }
        catch( IOException e )
        {
            UpgradePlugin.log( e );
        }
        
        try
        {
            writeFile( this.logFile, sw.toString() );
        }
        catch( CoreException e )
        {
            UpgradePlugin.log( e );
        }
    }

}
