/******************************************************************************
 * Copyright (c) 2016 Oracle
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/

package oracle.eclipse.tools.common.upgrade.internal;

import static oracle.eclipse.tools.common.upgrade.internal.UpgradePlugin.PROJECT_UPGRADE_PARTICIPANTS_EXTENSION_POINT_ID;
import static oracle.eclipse.tools.common.upgrade.internal.UpgradePlugin.WORKSPACE_UPGRADE_PARTICIPANTS_EXTENSION_POINT_ID;
import static oracle.eclipse.tools.common.upgrade.internal.UpgradePlugin.log;
import static oracle.eclipse.tools.util.EclipseThreadingUtil.waitForResourceChangesToCease;
import static oracle.eclipse.tools.util.PluginUtil.instantiate;
import static oracle.eclipse.tools.util.PluginUtil.reportMissingAttribute;

import java.lang.ref.SoftReference;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.osgi.util.NLS;
import org.eclipse.sapphire.util.MutableReference;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.PlatformUI;

import oracle.eclipse.tools.common.upgrade.IProjectUpgradeParticipant;
import oracle.eclipse.tools.common.upgrade.IProjectUpgradeParticipant.Action;
import oracle.eclipse.tools.common.upgrade.IWorkspaceUpgradeParticipant;
import oracle.eclipse.tools.util.PluginUtil;
import oracle.eclipse.tools.util.ProjectUtil;

/**
 * @author <a href="mailto:konstantin.komissarchik@oracle.com">Konstantin Komissarchik</a>
 */

public final class UpgradeFrameworkImpl 
{
    private static final String EL_PARTICIPANT = "participant"; //$NON-NLS-1$
    private static final String ATTR_ID = "id"; //$NON-NLS-1$
    private static final String ATTR_CLASS = "class"; //$NON-NLS-1$
    private static final String ATTR_REQUIRES = "requires"; //$NON-NLS-1$
    private static final String ATTR_PRIORITY = "priority"; //$NON-NLS-1$
    private static final int DEFAULT_PRIORITY = 0;
    private static final String PARTICIPANT_ID_SEPARATOR = ";"; //$NON-NLS-1$

    private final Map<String,SoftReference<ProjectUpgradeLog>> projectUpgradeLogs;
    private List<Set<Extension<IProjectUpgradeParticipant>>> projectUpgradeParticipants;
    private List<Set<Extension<IWorkspaceUpgradeParticipant>>> workspaceUpgradeParticipants;
    
    private final Object lock;
    private CheckerThread checkerThread;

    public UpgradeFrameworkImpl()
    {
        this.projectUpgradeLogs = new HashMap<String,SoftReference<ProjectUpgradeLog>>();
        this.lock = new Object();
        this.checkerThread = null;
    }
    
    public void initialize()
    {
        // Read the extensions.
        
        this.projectUpgradeParticipants
            = readExtensions( PROJECT_UPGRADE_PARTICIPANTS_EXTENSION_POINT_ID, 
                              IProjectUpgradeParticipant.class );
        
        this.workspaceUpgradeParticipants
            = readExtensions( WORKSPACE_UPGRADE_PARTICIPANTS_EXTENSION_POINT_ID, 
                              IWorkspaceUpgradeParticipant.class );
        
        // Register the resource change listener. It will listen for project
        // additions.
        
        final ResourceChangeListener listener = new ResourceChangeListener();
        final IWorkspace ws = ResourcesPlugin.getWorkspace();
        ws.addResourceChangeListener( listener, IResourceChangeEvent.POST_CHANGE );
        
        // Upgrade the workspace metadata, if necessary.
        
        boolean promptedForWorkspaceUpgrade = false;
        
        if( Tracing.isTracingEnabled( Tracing.Flag.WORKSPACE_UPGRADE ) )
        {
            Tracing.output( "Starting workspace upgrade." ); //$NON-NLS-1$
        }
        
        boolean restart = false;
        
        for( Set<Extension<IWorkspaceUpgradeParticipant>> bucket : getWorkspaceUpgradeParticipants() )
        {
            for( Extension<IWorkspaceUpgradeParticipant> pext : bucket )
            {
                final IWorkspaceUpgradeParticipant participant = pext.getParticipantInstance();
                
                try
                {
                    if( Tracing.isTracingEnabled( Tracing.Flag.WORKSPACE_UPGRADE ) )
                    {
                        final StringBuilder msg = new StringBuilder();
                        
                        msg.append( "CHECKING " ); //$NON-NLS-1$
                        msg.append( participant.getClass().getName() );
                        msg.append( " (" ); //$NON-NLS-1$
                        msg.append( pext.getPluginId() );
                        msg.append( ')' );
                        
                        Tracing.output( msg.toString() );
                    }
                    
                    if( participant.check() )
                    {
                        if( ! promptedForWorkspaceUpgrade )
                        {
                            if( ! promptForWorkspaceUpgrade() )
                            {
                                // User does not want to go through with workspace upgrade, so
                                // abort at this point.
                                
                                return;
                            }
                            
                            promptedForWorkspaceUpgrade = true;
                        }
                        
                        if( Tracing.isTracingEnabled( Tracing.Flag.WORKSPACE_UPGRADE ) )
                        {
                            final StringBuilder msg = new StringBuilder();
                            
                            msg.append( "EXECUTING " ); //$NON-NLS-1$
                            msg.append( participant.getClass().getName() );
                            msg.append( " (" ); //$NON-NLS-1$
                            msg.append( pext.getPluginId() );
                            msg.append( ')' );
                            
                            Tracing.output( msg.toString() );
                        }
                        
                        final boolean ready = participant.execute( new NullProgressMonitor() );
                        
                        restart = restart || ! ready;
                    }
                }
                catch( final Exception e )
                {
                    log( e );
                }
            }
        }
        
        if( Tracing.isTracingEnabled( Tracing.Flag.WORKSPACE_UPGRADE ) )
        {
            Tracing.output( "Workspace upgrade completed." ); //$NON-NLS-1$
        }
        
        if( restart )
        {
            runOnUIThread
            (
                new Runnable() 
                {
                    public void run()
                    {
                        showInformationDialog( Resources.restartRequiredDialogTitle, Resources.restartRequiredDialogMessage );
                        PlatformUI.getWorkbench().restart( true );
                    }
                }
            );
            
            return;
        }
        
        // Check whether the projects already in the workspace need to be upgraded.
        
        checkProjects( Arrays.asList( ws.getRoot().getProjects() ) );
    }
    
    public ProjectUpgradeLog getProjectUpgradeLog( final IProject project )
    {
        synchronized( this.projectUpgradeLogs )
        {
            // Clean the cache.
            
            final IWorkspaceRoot wsroot = ResourcesPlugin.getWorkspace().getRoot();
            final List<String> toDelete = new ArrayList<String>();
            
            for( Map.Entry<String,SoftReference<ProjectUpgradeLog>> entry
                 : this.projectUpgradeLogs.entrySet() )
            {
                final String projectName = entry.getKey();
                
                if( ! wsroot.getProject( projectName ).isAccessible() || 
                    entry.getValue().get() == null )
                {
                    toDelete.add( projectName );
                }
            }
            
            for( String projectName : toDelete )
            {
                this.projectUpgradeLogs.remove( projectName );
            }
            
            // Get the upgrade log for the requested project.
            
            SoftReference<ProjectUpgradeLog> softref 
                = this.projectUpgradeLogs.get( project.getName() );
            
            if( softref != null )
            {
                final ProjectUpgradeLog log = softref.get();
                
                if( log != null )
                {
                    log.refresh();
                    return log;
                }
            }
            
            final ProjectUpgradeLog log = new ProjectUpgradeLog( project );
            softref = new SoftReference<ProjectUpgradeLog>( log );
            this.projectUpgradeLogs.put( project.getName(), softref );
            
            return log;
        }
    }
    
    private List<Set<Extension<IProjectUpgradeParticipant>>> getProjectUpgradeParticipants()
    {
        return this.projectUpgradeParticipants;
    }

    private List<Set<Extension<IWorkspaceUpgradeParticipant>>> getWorkspaceUpgradeParticipants()
    {
        return this.workspaceUpgradeParticipants;
    }
    
    private void checkProjects( final Collection<IProject> projects )
    {
        if( projects != null && ! projects.isEmpty() )
        {
            synchronized( this.lock )
            {
                if( this.checkerThread == null )
                {
                    this.checkerThread = new CheckerThread( projects );
                    this.checkerThread.start();
                }
                else
                {
                    this.checkerThread.addProjects( projects );
                }
            }
        }
    }
    
    private static boolean promptForWorkspaceUpgrade()
    {
        if( showYesNoDialog( Resources.upgradeWorkspaceDialogTitle, Resources.upgradeWorkspaceDialogMessage ) )
        {
            return true;
        }
        else
        {
            showWarningDialog( Resources.upgradeCancelWarningDialogTitle, Resources.upgradeCancelWarningDialogMessage );
            
            return false;
        }
    }
    
    private static <T> List<Set<Extension<T>>> readExtensions( final String extensionPointId,
                                                               final Class<T> participantInterfaceClass )
    {
        final Set<Extension<T>> participants = new HashSet<Extension<T>>();
        final IExtensionRegistry registry = Platform.getExtensionRegistry();
        
        final IExtensionPoint point 
            = registry.getExtensionPoint( UpgradePlugin.SYMBOLIC_NAME, extensionPointId ); 
        
        if( point == null )
        {
            throw new IllegalStateException();
        }
        
        // Read the extensions.
        
        final Map<Extension,Set<String>> requiredParticipantsMap
            = new HashMap<Extension,Set<String>>();
        
        for( IExtension extension : point.getExtensions() )
        {
            final String pluginId = extension.getContributor().getName();
            final String pluginVersion = PluginUtil.getPluginVersion( pluginId );
            
            for( IConfigurationElement el : extension.getConfigurationElements() )
            {
                final String elName = el.getName();
                
                if( elName.equals( EL_PARTICIPANT ) )
                {
                    final String clname = el.getAttribute( ATTR_CLASS );
                    
                    if( clname == null )
                    {
                        reportMissingAttribute( el, ATTR_CLASS );
                    }
                    
                    String id = el.getAttribute( ATTR_ID );
                    
                    if( id == null )
                    {
                        id = clname;
                    }
                    
                    final String priorityStr = el.getAttribute(ATTR_PRIORITY);
                    int priority = DEFAULT_PRIORITY;
                    if (priorityStr != null) {
                    	try {
                    		priority = Integer.parseInt(priorityStr);
                    	}
                    	catch (NumberFormatException e) {
                            final String msg = NLS.bind( Resources.invalidPriority, id );
                            UpgradePlugin.logError( msg );
                    	}
                    }
                    
                    final String requires = el.getAttribute( ATTR_REQUIRES );
                    Set<String> requiredParticipants = null;
                    
                    if( requires != null )
                    {
                        for( String x : requires.split( PARTICIPANT_ID_SEPARATOR ) )
                        {
                            x = x.trim();
                            
                            if( x.length() > 0 )
                            {
                                if( requiredParticipants == null )
                                {
                                    requiredParticipants = new HashSet<String>();
                                }
                                
                                requiredParticipants.add( x );
                            }
                        }
                    }
                    
                    if( requiredParticipants == null )
                    {
                        requiredParticipants = Collections.emptySet();
                    }
                    
                    final T participant 
                        = instantiate( pluginId, clname, participantInterfaceClass );
                    
                    if( participant != null )
                    {
                        final Extension<T> pext 
                            = new Extension<T>( pluginId, pluginVersion, id, participant, priority );
                        
                        participants.add( pext );
                        requiredParticipantsMap.put( pext, requiredParticipants );
                    }
                    
                    
                }
            }
        }
        
        // Resolve the required participants id. Remove any participants that require
        // unresolvable participants.
        
        final Set<Extension> toRemove = new HashSet<Extension>();
        
        for( Extension<T> pinfo : participants )
        {
            final Set<String> requiredParticipants = requiredParticipantsMap.get( pinfo );
            
            for( String id : requiredParticipants )
            {
                final Extension<T> requiredParticipant = find( participants, id );
                
                if( requiredParticipant == null )
                {
                    toRemove.add( pinfo );
                    
                    final String msg 
                        = NLS.bind( Resources.participantDoesNotExist, id, pinfo.getId() );
                    
                    UpgradePlugin.logError( msg );
                }
                else
                {
                    pinfo.addRequiredParticipant( requiredParticipant );
                }
            }
        }
        
        boolean makeAnotherPass = true;
        
        while( makeAnotherPass )
        {
            makeAnotherPass = false;
            
            for( Extension<T> pinfo : participants )
            {
                if( toRemove.contains( pinfo ) )
                {
                    continue;
                }
                
                for( Extension<T> x : pinfo.getRequiredParticipants() )
                {
                    if( toRemove.contains( x ) )
                    {
                        toRemove.add( pinfo );
                        makeAnotherPass = true;
                        break;
                    }
                }
            }
        }
        
        for( Extension pinfo : toRemove )
        {
            participants.remove( pinfo );
            
            final String msg = NLS.bind( Resources.unsatisfiedDependencies, pinfo.getId() );
            UpgradePlugin.logError( msg );
        }
        
        // Sort the participants into buckets based on dependency order and return to the caller.
        
        return sort( participants );
    }
    
    private static <T> Extension<T> find( final Set<Extension<T>> participants,
                                          final String id )
    {
        for( Extension<T> pext : participants )
        {
            if( pext.getId().equals( id ) )
            {
                return pext;
            }
        }
        
        return null;
    }
    
    /**
     * Sorts the provided set of extensions into a list of buckets based on the dependencies and priority
     * specified by the extensions. The sort guarantees that any extensions required by 
     * extension A, or has a smaller priority value than extension A are placed in buckets prior to the bucket 
     * that the extension A is placed into.  If extenion A requires an extension with a higher priority tban A, 
     * it will be assumed that A has the same priority as the required extenison.  
     * 
     * <p>This bucketed sort is used as an optimization in lue of executing each extension (upgrade
     * participant) one at a time. Since upgrade participants modify files, a call to 
     * <code>IWorkspace.validateEdit()</code> is made which may cause prompts for the user if files
     * are read-only or under a source control system. Bucketing allows unrelated upgrade
     * participants to be grouped together so that calls to <code>IWorkspace.validateEdit()</code>
     * can be batched, thus reducing the number of prompts for the user.</p>
     * 
     * @param <T> the type of the extension
     * @param extensions the set of extensions to sort
     * @return the extensions sorted into a list of buckets
     */
    
    private static <T> List<Set<Extension<T>>> sort( final Set<Extension<T>> extensions )
    {
        final Set<Extension<T>> participants = new HashSet<Extension<T>>( extensions );
        final List<Set<Extension<T>>> sorted = new ArrayList<Set<Extension<T>>>();
        
        while( ! participants.isEmpty() )
        {
            final Set<Extension<T>> bucket = new HashSet<Extension<T>>();
            
            for( Extension<T> pinfo : participants )
            {
                boolean foundAllRequired = true;
                
                for( Extension<T> required : pinfo.getRequiredParticipants() )
                {
                    if( ! contains( sorted, required ) )
                    {
                        foundAllRequired = false;
                        break;
                    }
                }
                
                if( foundAllRequired )
                {
                	if (bucket.isEmpty()) {
                		bucket.add( pinfo );
                	}
                	else {
                		//Add the extension to the bucket if the bucket doesn't contain 
                		//any extensions with a lower priority value.  Any extensions in 
                		//the bucket with a higher priority value would be removed from 
                		//the bucket.
                		boolean addToBucket = true;

                		Iterator< Extension<T> > iterator = bucket.iterator();
                		while (iterator.hasNext()){
                			Extension<T> i = iterator.next();
                			if (i.getPriority() < pinfo.getPriority()) {
                				addToBucket = false;
                				break;
                			}
                			else if (i.getPriority() > pinfo.getPriority()) {
                				iterator.remove();
                			}
                		}
                		if (addToBucket) {
                			bucket.add( pinfo );
                		}
                	}
                }
            }
            
            if( ! bucket.isEmpty() )
            {
                sorted.add( bucket );
                participants.removeAll( bucket );
            }
            else
            {
                // Found a cycle. All of the participants still in the original
                // list are part of the cycle. Log the problem and disable them all.
                
                final StringBuilder buf = new StringBuilder();
                
                for( Extension pinfo : participants )
                {
                    if( buf.length() > 0 ) buf.append( ", " ); //$NON-NLS-1$
                    buf.append( '"' );
                    buf.append( pinfo.getId() );
                    buf.append( '"' );
                }
                
                final String msg = NLS.bind( Resources.cycleDetected, buf.toString() );
                UpgradePlugin.logError( msg );
                
                participants.clear();
            }
        }
        
        traceSortResults( extensions, sorted );

        return sorted;
    }
    
    private static <T>void traceSortResults( final Set<Extension<T>> input,
                                             final List<Set<Extension<T>>> output )
    {
        if( Tracing.isTracingEnabled( Tracing.Flag.INITIALIZATION ) )
        {
            final StringBuilder msg = new StringBuilder();
            
            msg.append( "Sorting upgrade participants: \n\n" ); //$NON-NLS-1$
            msg.append( "  Input:\n\n" ); //$NON-NLS-1$
            
            for( Extension<T> extension : input )
            {
                msg.append( "    " ); //$NON-NLS-1$
                msg.append( extension.getParticipantInstance().getClass().getName() );
                msg.append( " (" ); //$NON-NLS-1$
                msg.append( extension.getPluginId() );
                msg.append( ")\n" ); //$NON-NLS-1$
            }
            
            msg.append( "\n  Output:\n\n" ); //$NON-NLS-1$
            
            int counter = 1;
            
            for( Set<Extension<T>> bucket : output )
            {
                if( counter > 1 )
                {
                    msg.append( '\n' );
                }
                
                msg.append( "    Bucket " ); //$NON-NLS-1$
                msg.append( counter );
                msg.append( ":\n\n" ); //$NON-NLS-1$
                
                for( Extension<T> extension : bucket )
                {
                    msg.append( "      " ); //$NON-NLS-1$
                    msg.append( extension.getParticipantInstance().getClass().getName() );
                    msg.append( " (" ); //$NON-NLS-1$
                    msg.append( extension.getPluginId() );
                    msg.append( ")\n" ); //$NON-NLS-1$
                }
                
                counter++;
            }
            
            Tracing.output( msg.toString() );
        }
    }

    private static <T> boolean contains( final List<Set<Extension<T>>> extensions,
                                         final Extension<T> extension )
    {
        for( Set<Extension<T>> set : extensions )
        {
            if( set.contains( extension ) )
            {
                return true;
            }
        }
        
        return false;
    }
    
    private static <T> int size( final List<Set<Extension<T>>> extensions )
    {
        int size = 0;
        
        for( Set<Extension<T>> set : extensions )
        {
            size += set.size();
        }
        
        return size;
    }
    
    private static IStatus createErrorStatus( final Throwable e )
    {
        return new Status( IStatus.ERROR, UpgradePlugin.SYMBOLIC_NAME, 0, Resources.unexpectedException, e );
    }
    
    private static boolean hasModalShell( final Display display )
    {
        if( display.isDisposed() )
        {
            return false;
        }
        else
        {
            final MutableReference<Boolean> result = new MutableReference<Boolean>();
            
            display.syncExec
            (
                new Runnable()
                {
                    public void run()
                    {
                        result.set( Boolean.FALSE );
                        
                        for( Shell shell : display.getShells() )
                        {
                            if( ! shell.isDisposed() && shell.isVisible() &&
                                ( shell.getStyle() & SWT.APPLICATION_MODAL ) != 0 )
                            {
                                result.set( Boolean.TRUE );
                                break;
                            }
                        }
                    }
                }
            );
            
            return result.get().booleanValue();
        }
    }
    
    private static void waitOnModalShells()
    {
        final Display display = PlatformUI.getWorkbench().getDisplay();
        
        while( hasModalShell( display ) )
        {
            try
            {
                Thread.sleep( 1000 );
            }
            catch( InterruptedException e ) {}
        }
    }
    
    private static boolean showYesNoDialog( final String title, final String message )
    {
        final MutableReference<Boolean> result = new MutableReference<Boolean>( false );
        
        final Runnable runnable = new Runnable() 
        {
            public void run()
            {
                final Shell shell = Display.getCurrent().getActiveShell();
                result.set( MessageDialog.openQuestion( shell, title, message ) );
            }
        };
        
        runOnUIThread( runnable );
        
        return result.get();
    }
    
    private static void showWarningDialog( final String title, final String message )
    {
        final Runnable runnable = new Runnable() 
        {
            public void run()
            {
                final Shell shell = Display.getCurrent().getActiveShell();
                MessageDialog.openWarning( shell, title, message );
            }
        };
        
        runOnUIThread( runnable );
    }
    
    private static void showInformationDialog( final String title, final String message )
    {
        final Runnable runnable = new Runnable() 
        {
            public void run()
            {
                final Shell shell = Display.getCurrent().getActiveShell();
                MessageDialog.openInformation( shell, title, message );
            }
        };
        
        runOnUIThread( runnable );
    }
    
    private static void runOnUIThread(final Runnable runnable)
    {
        final Display display = PlatformUI.getWorkbench().getDisplay();
        
        if( Display.findDisplay( Thread.currentThread() ) == display )
        {
            runnable.run();
        }
        else
        {
            display.syncExec( runnable );
        }
    }
    
    private static final class Extension<T>
    {
        private final String pluginId;
        private final String pluginVersion;
        private final String id;
        private final T participant;
        private final int priority;
        private final Set<Extension<T>> requiredParticipants;
        private final Set<Extension<T>> requiredParticipantsReadOnly;
        
        public Extension( final String pluginId,
                          final String pluginVersion,
                          final String id,
                          final T participant,
                          final int priority)
        {
            this.pluginId = pluginId;
            this.pluginVersion = ( pluginVersion != null ? pluginVersion : "N/A" ); //$NON-NLS-1$
            this.id = id;
            this.participant = participant;
            this.priority = priority;
            
            this.requiredParticipants = new HashSet<Extension<T>>();
            
            this.requiredParticipantsReadOnly 
                = Collections.unmodifiableSet( this.requiredParticipants );
        }
        
        public String getPluginId()
        {
            return this.pluginId;
        }
        
        public String getPluginVersion()
        {
            return this.pluginVersion;
        }
        
        public String getId()
        {
            return this.id;
        }
        
        public T getParticipantInstance()
        {
            return this.participant;
        }
        
        public int getPriority()
        {
        	return this.priority;
        }
        
        public Set<Extension<T>> getRequiredParticipants()
        {
            return this.requiredParticipantsReadOnly;
        }
        
        public void addRequiredParticipant( final Extension<T> pdesc )
        {
            this.requiredParticipants.add( pdesc );
        }
        
        public String toString()
        {
            return this.id;
        }
    }
    
    private final class ResourceChangeListener
    
        implements IResourceChangeListener
        
    {
        public void resourceChanged( final IResourceChangeEvent event ) 
        {
            final IResourceDelta delta = event.getDelta();
            
            if( delta == null )
            {
                return;
            }
            
            Set<IProject> projectsToCheck = null;
            
            for( IResourceDelta d : delta.getAffectedChildren() )
            {
                final IProject pj = (IProject) d.getResource();
                final int kind = d.getKind();
                
                boolean addProjectToQueue = false;
                
                if( kind == IResourceDelta.ADDED )
                {
                    addProjectToQueue = true;
                }
                else if( kind == IResourceDelta.CHANGED && 
                        ( d.getFlags() & IResourceDelta.OPEN ) != 0 )
                {
                    // This delta is generated both for open and close
                    // events, so we need an extra check to handle only
                    // the open events.
                    
                    if( pj.isAccessible() )
                    {
                        addProjectToQueue = true;
                    }
                }
                
                if( addProjectToQueue )
                {
                    if( projectsToCheck == null )
                    {
                        projectsToCheck = new HashSet<IProject>();
                    }
                    
                    projectsToCheck.add( pj );
                }
            }
            
            checkProjects( projectsToCheck );
        }
    }
    
    private final class CheckerThread
    
        extends Thread
        
    {
        private final Set<IProject> projectsToCheck;
        
        public CheckerThread( final Collection<IProject> projectsToCheck )
        {
            this.projectsToCheck = new HashSet<IProject>();
            this.projectsToCheck.addAll( projectsToCheck );
        }
        
        @Override
        public void run()
        {
            // Wait for all change activity to settle down before checking projects. This
            // accomplishes two things. First, it makes it far less likely that we would be checking
            // incomplete projects. Second, it allows additional projects to trickle into this
            // upgrade batch in the case of a multi-project operation such as importing several 
            // projects.
            
            waitOnModalShells();
            waitForResourceChangesToCease( 5 );
            waitOnModalShells();
            
            synchronized( UpgradeFrameworkImpl.this.lock )
            {
                UpgradeFrameworkImpl.this.checkerThread = null;
            }
            
            // Check the accumulated project set for projects that need to be upgraded and schedule
            // an upgrade job if necessary.
            
            final Set<IProject> projectsToUpgrade = new HashSet<IProject>();
            
            for( IProject project : this.projectsToCheck )
            {
                if( checkProject( project ) )
                {
                    projectsToUpgrade.add( project );
                }
            }
            
            //CR346961, CR352835
            
            if( ! projectsToUpgrade.isEmpty() )
            {
                final UpgradeJob upgradeJob = new UpgradeJob( projectsToUpgrade );
                upgradeJob.schedule();
            }
        }
        
        public void addProjects( final Collection<IProject> projects )
        {
            synchronized( UpgradeFrameworkImpl.this.lock )
            {
                this.projectsToCheck.addAll( projects );
            }
        }
        
        private boolean checkProject( final IProject project )
        {
            for( Set<Extension<IProjectUpgradeParticipant>> bucket : getProjectUpgradeParticipants() )
            {
                for( Extension<IProjectUpgradeParticipant> participant : bucket )
                {
                    // Keep checking the project accessible status in case
                    // the project is closed while we are inside this method.
                    
                    if( ! project.isAccessible() )
                    {
                        return false;
                    }
                    
                    try
                    {
                        if( participant.getParticipantInstance().check( project ) )
                        {
                            return true;
                        }
                    }
                    catch( final Exception e )
                    {
                        log( e );
                    }
                }
            }
            
            return false;
        }
    }
    
    private final class UpgradeJob
    
        extends Job
        
    {
        private final Set<IProject> projects;
        
        public UpgradeJob( final Set<IProject> projects )
        {
            super( Resources.upgradeJobTitle );
            setRule( ResourcesPlugin.getWorkspace().getRoot() );
            this.projects = projects;
        }
        
        @Override
        protected IStatus run( final IProgressMonitor monitor ) 
        {
            final MutableReference<IStatus> result = new MutableReference<IStatus>();
            
            final IWorkspaceRunnable wr = new IWorkspaceRunnable()
            {
                public void run( final IProgressMonitor monitor )
                {
                    result.set( executeUpgrade( monitor ) );
                }
            };
            
            IStatus status;
            
            try
            {
                ResourcesPlugin.getWorkspace().run( wr, monitor );
                status = result.get();
            }
            catch( CoreException e )
            {
                log( e );
                status = e.getStatus();
            }
            
            return status;
        }

        private IStatus executeUpgrade( final IProgressMonitor monitor ) 
        {
            final List<Set<Extension<IProjectUpgradeParticipant>>> buckets 
                = getProjectUpgradeParticipants();
            
            final Set<IStatus> problems = new HashSet<IStatus>();
            final int workItemsCount = this.projects.size() * size( buckets ) + 4;
            
            monitor.beginTask( Resources.upgradeTaskLabel, workItemsCount );
            
            IProgressMonitor wizardProgressMonitor = null;
            
            try
            {
                // Open the upgrade wizard and block until the user presses
                // the Finish or the Cancel button.
                
                final UpgradeWizard wizard = new UpgradeWizard( this.projects );
                
                wizard.show();
                
                if( wizard.isCanceled() )
                {
                    return Status.CANCEL_STATUS;
                }
                
                wizardProgressMonitor = wizard.getProgressMonitor();
                wizardProgressMonitor.beginTask( Resources.upgradeTaskLabel, workItemsCount );
                
                final IProgressMonitor combinedMonitor
                    = new BroadcastProgressMonitor( monitor, wizardProgressMonitor );
                
                // Make sure the project resources are in sync with the file system,
                // otherwise the upgrade might fail.
                
                for( IProject project : this.projects )
                {
                    try
                    {
                        project.refreshLocal( IResource.DEPTH_INFINITE, null );
                    }
                    catch( CoreException e )
                    {
                        return e.getStatus();
                    }
                }
                
                combinedMonitor.worked( 1 );
                
                // Process the upgrade participants one bucket at a time.
                
                for( Set<Extension<IProjectUpgradeParticipant>> bucket : buckets )
                {
                    try
                    {
                        processUpgradeParticipants( bucket, combinedMonitor, problems );
                    }
                    catch( AbortUpgradeException e )
                    {
                        return Status.OK_STATUS;
                    }
                }
                
                // Schedule a job to re-build the upgraded projects.
                
                ProjectUtil.scheduleCleanBuild( this.projects );
                combinedMonitor.worked( 1 );
            }
            finally
            {
                monitor.done();
                
                if( wizardProgressMonitor != null )
                {
                    wizardProgressMonitor.done();
                }
            }
            
            final IStatus status;
            
            if( problems.isEmpty() )
            {
                status = Status.OK_STATUS;
            }
            else
            {
                status = new MultiStatus( UpgradePlugin.SYMBOLIC_NAME, 0,
                                          problems.toArray( new IStatus[ problems.size() ] ), 
                                          Resources.failedWhileUpgrading,
                                          null );
            }
            
            return status;
        }
        
        private void processUpgradeParticipants( final Set<Extension<IProjectUpgradeParticipant>> bucket,
                                                 final IProgressMonitor combinedMonitor,
                                                 final Set<IStatus> problems )
        
            throws AbortUpgradeException
            
        {
            // Prepare the participants. Gather the set of files that will be edited.
            
            final Map<IProject,Set<Extension<IProjectUpgradeParticipant>>> activeParticipants
                = new HashMap<IProject,Set<Extension<IProjectUpgradeParticipant>>>();
            
            final Set<IFile> filesToBeEdited = new HashSet<IFile>();
            
            for( IProject project : this.projects )
            {
                filesToBeEdited.add( getProjectUpgradeLog( project ).getLogFile() );
                
                final Set<Extension<IProjectUpgradeParticipant>> activeParticipantsForProject
                    = new HashSet<Extension<IProjectUpgradeParticipant>>();
                
                activeParticipants.put( project, activeParticipantsForProject );
                
                for( Extension<IProjectUpgradeParticipant> participant : bucket )
                {
                    try
                    {
                        final Set<Action> participantActions
                            = participant.getParticipantInstance().prepare( project );
                        
                        if( ! participantActions.isEmpty() )
                        {
                            activeParticipantsForProject.add( participant );
                            
                            for( Action action : participantActions )
                            {
                                if(( action.getType() == Action.Type.EDIT) ||
                                   ( action.getType() == Action.Type.DELETE))
                                {
                                    filesToBeEdited.add( action.getFile() );
                                }
                            }
                        }
                    }
                    catch( CoreException e )
                    {
                        problems.add( e.getStatus() );
                    }
                    catch( Throwable e )
                    {
                        problems.add( createErrorStatus( e ) );
                    }
                }
            }
            
            combinedMonitor.worked( 1 );
            
            // Validate that the files to be edited can indeed be edited.
   
            final Display display = PlatformUI.getWorkbench().getDisplay();
            final IWorkspace ws = ResourcesPlugin.getWorkspace();
            
            final IFile[] filesToBeEditedArray
                = filesToBeEdited.toArray( new IFile[ filesToBeEdited.size() ] );
            
            final MutableReference<IStatus> validateEditResult
                = new MutableReference<IStatus>();
            
            display.syncExec
            (
                new Runnable()
                {
                    public void run() 
                    {
                        final IStatus st 
                            = ws.validateEdit( filesToBeEditedArray, display.getActiveShell() );
                        
                        validateEditResult.set( st ); 
                    }
                }
            );
            
            if( validateEditResult.get().getSeverity() == IStatus.ERROR )
            {
                throw new AbortUpgradeException();
            }
            
            combinedMonitor.worked( 1 );
            
            // Execute the prepared upgrade participants.
            
            for( IProject project : this.projects )
            {
                final Set<Extension<IProjectUpgradeParticipant>> activeParticipantsForProject
                    = activeParticipants.get( project );
                
                final ProjectUpgradeLog upgradeLog = getProjectUpgradeLog( project );
                
                for( Extension<IProjectUpgradeParticipant> participant : activeParticipantsForProject )
                {
                    try
                    {
                        final NullProgressMonitor npm = new NullProgressMonitor();
                        final IProjectUpgradeParticipant p = participant.getParticipantInstance();
                        
                        p.execute( project, npm );
                        
                        upgradeLog.addEntry( participant.getPluginId(), 
                                             participant.getPluginVersion(), 
                                             p.getClass().getName() );
                    }
                    catch( CoreException e )
                    {
                        problems.add( e.getStatus() );
                    }
                    catch( Throwable e )
                    {
                        problems.add( createErrorStatus( e ) );
                    }
                    
                    combinedMonitor.worked( 1 );
                }
                
                for( int i = 0, n = bucket.size() - activeParticipantsForProject.size(); 
                     i < n; i++ )
                {
                    combinedMonitor.worked( 1 );
                }
            }
        }
    }
    
    private static final class AbortUpgradeException
    
        extends Exception
        
    {
        private static final long serialVersionUID = 1L;
    }
    
    private static final class BroadcastProgressMonitor
    
        implements IProgressMonitor
        
    {
        private final IProgressMonitor monitor1;
        private final IProgressMonitor monitor2;
        
        public BroadcastProgressMonitor( final IProgressMonitor monitor1,
                                         final IProgressMonitor monitor2 )
        {
            this.monitor1 = monitor1;
            this.monitor2 = monitor2;
        }

        public void beginTask( final String name, 
                               final int totalWork ) 
        {
            // Do not forward this event. The monitors are assumed to have been
            // started externally.
        }

        public void done() 
        {
            this.monitor1.done();
            this.monitor2.done();
        }

        public void internalWorked( final double work ) 
        {
            this.monitor1.internalWorked( work );
            this.monitor2.internalWorked( work );
        }

        public boolean isCanceled() 
        {
            return this.monitor1.isCanceled() || this.monitor2.isCanceled();
        }

        public void setCanceled( final boolean value ) 
        {
            this.monitor1.setCanceled( value );
            this.monitor2.setCanceled( value );
        }

        public void setTaskName( final String name ) 
        {
            this.monitor1.setTaskName( name );
            this.monitor2.setTaskName( name );
        }

        public void subTask( final String name ) 
        {
            this.monitor1.subTask( name );
            this.monitor2.subTask( name );
        }

        public void worked( final int work ) 
        {
            this.monitor1.worked( work );
            this.monitor2.worked( work );
        }
    }

    private static final class Resources extends NLS
    {
        public static String unexpectedException;
        public static String upgradeJobTitle;
        public static String upgradeTaskLabel;
        public static String failedWhileUpgrading;
        public static String participantDoesNotExist;
        public static String unsatisfiedDependencies;
        public static String cycleDetected;
        public static String upgradeWorkspaceDialogTitle;
        public static String upgradeWorkspaceDialogMessage;
        public static String upgradeCancelWarningDialogTitle;
        public static String upgradeCancelWarningDialogMessage;
        public static String restartRequiredDialogTitle;
        public static String restartRequiredDialogMessage;
        public static String invalidPriority;
        
        static
        {
            initializeMessages( UpgradeFrameworkImpl.class.getName(), Resources.class );
        }
    }
    
}
