/******************************************************************************
 * Copyright (c) 2016 Oracle
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/

package oracle.eclipse.tools.common.upgrade.internal;

import java.lang.reflect.InvocationTargetException;
import java.util.Set;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.ProgressMonitorWrapper;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.wizard.Wizard;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.PlatformUI;

/**
 * @author <a href="mailto:konstantin.komissarchik@oracle.com">Konstantin Komissarchik</a>
 */

public final class UpgradeWizard 

    extends Wizard
    
{
    private final Set<IProject> projects;
    private boolean isCanceled;
    private IProgressMonitor monitor;
    
    public UpgradeWizard( final Set<IProject> projects )
    {
        this.projects = projects;
        this.isCanceled = true;
        this.monitor = null;

        setNeedsProgressMonitor( true );
        setWindowTitle( Resources.windowTitle );
    }
    
    public void addPages() 
    {
        final UpgradeWizardPage page = new UpgradeWizardPage( this.projects );
        addPage( page );
    }

    @Override
    public boolean performFinish() 
    {
        final IRunnableWithProgress op = new IRunnableWithProgress()
        {
            public void run( final IProgressMonitor monitor )
            {
                final UpgradeWizard wizard = UpgradeWizard.this;
                
                synchronized( wizard )
                {
                    // The job that launched the wizard is waiting. We create a progress
                    // monitor that will notify us when the job calls the done method.
                    // We then notify the job thread that's waiting and start waiting
                    // ourselves. When we are notified we know that the operations has 
                    // completed and we can exit out of this method which will cause the 
                    // upgrade wizard to close.

                    wizard.isCanceled = false;
                    wizard.monitor = new NotifyWhenDoneProgressMonitor( monitor, wizard );
                    
                    wizard.notify();
                    
                    boolean keepWaiting = true;
                    
                    while( keepWaiting )
                    {
                        try
                        {
                            wizard.wait();
                            keepWaiting = false;
                        }
                        catch( InterruptedException e ) {}
                    }
                }
            }
        };
        
        try 
        {
            getContainer().run( true, false, op );
        }
        catch( InterruptedException e ) 
        {
            return false;
        } 
        catch( InvocationTargetException e ) 
        {
            throw new RuntimeException( e.getTargetException() );
        }
        
        return true;
    }
    
    @Override
    public synchronized boolean performCancel()
    {
        this.isCanceled = true;
        notify();
        
        MessageDialog.openWarning( getShell(), Resources.upgradeCancelWarningDialogTitle, Resources.upgradeCancelWarningMessage );
        
        return true;
    }
    
    public synchronized void show()
    {
        final Display display = PlatformUI.getWorkbench().getDisplay();
        
        display.asyncExec
        (
            new Runnable()
            {
                public void run() 
                {
                    final UpgradeWizard wizard = UpgradeWizard.this;
                    
                    try
                    {
                        final Shell shell = display.getActiveShell();
                        final WizardDialog dialog = new WizardDialog( shell, wizard );
                        
                        dialog.open();
                    }
                    finally
                    {
                        synchronized( wizard )
                        {
                            wizard.notify();
                        }
                    }
                }
            }
        );
        
        boolean keepWaiting = true;
        
        while( keepWaiting )
        {
            try
            {
                wait();
                keepWaiting = false;
            }
            catch( InterruptedException e ) {}
        }
    }
    
    public synchronized boolean isCanceled()
    {
        return this.isCanceled;
    }
    
    public synchronized IProgressMonitor getProgressMonitor()
    {
        return this.monitor;
    }
    
    private static final class NotifyWhenDoneProgressMonitor
    
        extends ProgressMonitorWrapper
        
    {
        private final Object monitor;
        
        public NotifyWhenDoneProgressMonitor( final IProgressMonitor progressMonitor,
                                                     final Object monitor ) 
        {
            super( progressMonitor );
            this.monitor = monitor;
        }
        
        public void done()
        {
            super.done();
            
            synchronized( this.monitor )
            {
                this.monitor.notify();
            }
        }
    }
    
    private static final class Resources
    
        extends NLS
        
    {
        public static String windowTitle;
        public static String upgradeCancelWarningDialogTitle;
        public static String upgradeCancelWarningMessage;
        
        static
        {
            initializeMessages( UpgradeWizard.class.getName(), 
                                Resources.class );
        }
    }
    
}
