/******************************************************************************
 * Copyright (c) 2016 Oracle
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/

package oracle.eclipse.tools.util;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.nio.channels.FileChannel;
import java.nio.charset.StandardCharsets;
import java.util.Comparator;
import java.util.regex.Pattern;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Plugin;
import org.eclipse.core.runtime.content.IContentDescription;
import org.eclipse.core.runtime.content.IContentType;
import org.eclipse.core.runtime.jobs.ISchedulingRule;
import org.eclipse.osgi.util.NLS;
import org.osgi.framework.Bundle;

import oracle.eclipse.tools.util.internal.UtilPlugin;

public final class EclipseFileUtil
{
	public static final boolean caseSensitive = Platform.OS_MACOSX.equals(Platform.getOS()) ? false : new java.io.File("a").compareTo(new java.io.File("A")) != 0; //$NON-NLS-1$ //$NON-NLS-2$


    /**
     * Copies a resource from a plugin to a destination in the workspace.
     *
     * @param pluginId the plugin id
     * @param pathInPlugin the path of the resource within the plugin
     * @param destination the destination path within the workspace
     */
    
    public static void copyFromPlugin( final String pluginId,
                                       final IPath pathInPlugin,
                                       final IFile destination )

        throws CoreException

    {
        copyFromPlugin( pluginId, pathInPlugin, destination.getLocation().toFile() );
    }
    
    /**
     * Copies a resource from a plugin to a destination in the workspace.
     *
     * @param bundle the plugin bundle
     * @param pathInPlugin the path of the resource within the plugin
     * @param destination the destination path within the workspace
     */

    public static void copyFromPlugin( final Bundle bundle,
                                       final IPath pathInPlugin,
                                       final IFile destination )

        throws CoreException

    {
        copyFromPlugin( bundle, pathInPlugin, destination.getLocation().toFile() );
    }
    
    /**
     * Copies a resource from a plugin to a destination on the file system.
     *
     * @param pluginId the plugin id
     * @param pathInPlugin the path of the resource within the plugin
     * @param destination the destination path within the workspace
     */
    
    public static void copyFromPlugin( final String pluginId,
                                       final IPath pathInPlugin,
                                       final File destination )
    
        throws CoreException
        
    {
        final Bundle bundle = Platform.getBundle( pluginId );
        copyFromPlugin( bundle, pathInPlugin, destination );
    }
    
    /**
     * Copies a resource from a plugin to a destination on the file system.
     *
     * @param bundle the plugin bundle
     * @param pathInPlugin the path of the resource within the plugin
     * @param destination the destination path within the workspace
     */

    public static void copyFromPlugin( final Bundle bundle,
                                       final IPath pathInPlugin,
                                       final File destination )
    
        throws CoreException
        
    {
        try( InputStream contents = readFileFromPlugin( bundle, pathInPlugin ) )
        {
            writeFile( destination, contents );
        }
        catch( final IOException e )
        {
        	throw new CoreException( UtilPlugin.createErrorStatus( e ) );
		}
    }
    
    public static InputStream readFileFromPlugin( final Bundle bundle,
                                                  final IPath pathInPlugin )
    
        throws CoreException
        
    {
        try
        {
            return FileLocator.openStream( bundle, pathInPlugin, false );
        }
        catch( final IOException e )
        {
            throw new CoreException( UtilPlugin.createErrorStatus( e ) );
        }
    }
    
    public static File getFileFromPlugin(Plugin plugin, String pluginRelativePath)
    {
        File file = null;
        
        // The plugin 
        Bundle bundle = plugin.getBundle(); 
        try
        {
            file = getFileFromPlugin(bundle, pluginRelativePath);
        }
        catch(IOException e)
        {
            UtilPlugin.log(e);
            assert(false) : e.getMessage();
        }
        
        return file;
    }
    
    public static File getFileFromPlugin(Bundle bundle, String pluginRelativePath)
        throws IOException
    {
        // The plugin 
        URL iconDir = FileLocator.find(bundle, new Path(pluginRelativePath), null);
        assert ( iconDir != null ) : "Couldn't find path in plugin bundle: " + pluginRelativePath; //$NON-NLS-1$
        
        URL localURL = FileLocator.toFileURL(iconDir);
        return new File(localURL.getPath());
    }

    public static void copyFileToProject(File from, IFile to) 
    		throws IOException, CoreException
    {
    	copyFile(from, to.getLocation().toFile());
    	to.getProject().refreshLocal( IResource.DEPTH_INFINITE, null );
    }
    
    public static void copyFile(File from, File to) throws IOException
    {
        assert(from != null);
        assert(to != null);
        
        // first create the "to" file if it doesn't already exist
        if(!to.exists())
        {
            File parentDir = to.getParentFile();

            if( ! parentDir.exists() && ! parentDir.mkdirs() )
            {
                throw new IOException();
            }

            to.createNewFile();
        }

        FileInputStream in = null;
        FileOutputStream out = null;
        try
        {
            in = new FileInputStream(from);
            try
            {
                out = new FileOutputStream(to);
                FileChannel ic = in.getChannel();
                try
                {
                    FileChannel oc = out.getChannel();
                    try
                    {
                        oc.transferFrom(ic, 0, from.length());
                    }
                    finally
                    {
                        if(oc != null)
                        {
                            oc.close();
                        }
                    }
                }
                finally
                {
                    if(ic != null)
                    {
                        ic.close();
                    }
                }
            }
            finally
            {
                if(out != null)
                {
                    out.close();
                }
            }
        }
        finally
        {
            if(in != null)
            {
                in.close();
            }
        }
    }
    
    /**
     * Copies the contents of the source File to the destination File. If 
     * the args are directories, performs a recurive copy. 
     * @param from Source file. If a directory, contents will be copied recursively.
     * @param to Destination file. If source is a directory, must be a directory.
     * @throws IOException Thrown if an error is encountered.
     */
    public static void copyFiles(final File from, final File to) throws IOException
    {
        copyFiles(from, to, (String) null);
    }
    
    /**
     * Copies the contents of the source File to the destination File. If 
     * the args are directories, performs a recurive copy. If the pattern regexp is
     * specified, only files that match the pattern will be copied.
     * @param from Source file. If a directory, contents will be copied recursively.
     * @param to Destination file. If source is a directory, must be a directory.
     * @param regexpPattern Optional regular expression pattern. If specified, 
     * source file names must match to be copied and empty directories will not be copied.
     * @return True if any files were copied.
     * @throws IOException Thrown if an error is encountered.
     */
    public static boolean copyFiles(final File from, final File to, final String regexp) throws IOException
    {
        assert(from != null);
        assert(to   != null);
        
        Pattern pattern = null;
        if (regexp != null) {
            pattern = Pattern.compile(regexp);
        }
        return copyFiles(from, to, pattern);
    }
     
    private static boolean copyFiles(final File from, final File to, final Pattern regexp) throws IOException
    {
        if(from.isDirectory())
        {
            if( ! to.exists() && ! to.mkdirs() )
            {
                throw new IOException();
            } 

            assert (to.isDirectory());
            
            final File[] fromChildren = from.listFiles();
        	boolean copiedOneChild = false;
            if(fromChildren != null)
            {
                for(File nextFrom : fromChildren)
                {
                    final File nextTo = new File(to, nextFrom.getName());
                    if (copyFiles(nextFrom, nextTo, regexp)) {
                    	copiedOneChild = true;
                    }
                }
            }
            if (regexp != null && !copiedOneChild) {
            	// did not copy any children so delete empty destination directory
            	if( ! to.delete() )
            	{
            	    throw new IOException();
            	}
            }

            return copiedOneChild;
        }
        else
        {
            if (regexp == null || regexp.matcher(from.getName()).matches()) {
            	copyFile(from, to);
            	return true;
            }
            return false;
        }
    }

    /**
     * Deletes the specified file. If a directory, recursively deletes all
     * contents.
     * @param file File or directory to delete.
     * @return True if successfully deleted.
     * @throws IOException Thrown if an error is encountered.
     */
    public static boolean deleteFile(final File file) throws IOException
    {
        assert (file != null);

        if(file.isDirectory())
        {
            final File[] children = file.listFiles();
            if (children != null) {
                for(File child : children)
                {
                    deleteFile(child);
                }
            }
        }
    	return file.delete();
    }
    
    /**
     * Determine if the given file's content type matches the given content
     * type id. If the file does not exist, or an error is encountered trying
     * to get its content type, this will return false.
     */
    public static boolean isContentType( IFile file, String contentTypeId )
    {
        assert ( file != null );
        ISchedulingRule rule = null;
        try {
            // make sure the file will not get deleted while we are inspecting it
            rule = ResourcesPlugin.getWorkspace().getRuleFactory().deleteRule( file );
            Platform.getJobManager().beginRule( rule, null );
            if ( !file.exists() )
                return false;
            
            IContentDescription contentDesc = null;
            try
            {
                contentDesc = file.getContentDescription();
            }
            catch ( CoreException e )
            {
                // based on the javadoc for getContentDescription, a number of things can go
                // wrong, all of which would mean that the file being looked at is not the given
                // content type. So if we catch a CoreException here, just make sure the
                // content description is null, and move on.
                contentDesc = null;
            }
            
            if (contentDesc != null)
            {
                IContentType contentType = contentDesc.getContentType();
                if (contentType != null)
                {
                    if ( contentType.getId().equals(contentTypeId) )
                        return true;
                }
            }
        }
        finally
        {
            if ( rule != null )
            {
                Platform.getJobManager().endRule( rule );
            }
        }

        return false;
    }
    
    public static IFile getWorkspaceFile( final File f )
    {
        final IWorkspace ws = ResourcesPlugin.getWorkspace();
        final IWorkspaceRoot wsroot = ws.getRoot();
        final IPath path = new Path( f.getAbsolutePath() );
        
        final IFile[] wsFiles = wsroot.findFilesForLocation( path );
        
        if( wsFiles.length > 0 )
        {
            return wsFiles[ 0 ];
        }
        
        return null;
    }

    public static IContainer getWorkspaceContainer( final File f )
    {
        final IWorkspace ws = ResourcesPlugin.getWorkspace();
        final IWorkspaceRoot wsroot = ws.getRoot();
        final IPath path = new Path( f.getAbsolutePath() );
        
        final IContainer[] wsContainers = wsroot.findContainersForLocation( path );
        
        if( wsContainers.length > 0 )
        {
            return wsContainers[ 0 ];
        }
        
        return null;
    }
    
    public static void mkdirs( final File f )
    
        throws CoreException
        
    {
        final IContainer wsContainer = getWorkspaceContainer( f );
        
        if( f.exists() )
        {
            if( f.isFile() )
            {
                final String msg
                    = NLS.bind( Resources.locationIsFile, 
                                f.getAbsolutePath() );
                
                throw new CoreException( UtilPlugin.createErrorStatus( msg ) );
            }
            else
            {
                // Make sure that the the folder is in the workspace.
                
                if( wsContainer != null )
                {
                    wsContainer.refreshLocal( IResource.DEPTH_ZERO, null );
                }
            }
        }
        else
        {
            mkdirs( f.getParentFile() );
            
            if( wsContainer != null )
            {
                // Should be a folder...
                
                final IFolder iFolder = (IFolder) wsContainer;
                iFolder.create( true, true, null );
            }
            else
            {
                final boolean isSuccessful = f.mkdir();
                
                if( ! isSuccessful )
                {
                    final String msg
                        = NLS.bind( Resources.failedToCreateDirectory, 
                                    f.getAbsolutePath() );
                    
                    throw new CoreException( UtilPlugin.createErrorStatus( msg ) );
                }
            }
        }
    }
    
    public static void writeFile( final IFile f,
                                  final String contents )
    
        throws CoreException
        
    {
        writeFile( f.getLocation().toFile(), contents );
    }
    
    public static void writeFile( final IFile f,
                                  final byte[] contents )
    
        throws CoreException
        
    {
        writeFile( f.getLocation().toFile(), contents );
    }

    public static void writeFile( final IFile f,
                                  final InputStream contents )
    
        throws CoreException
        
    {
        writeFile( f.getLocation().toFile(), contents );
    }

    public static void writeFile( final File f,
                                  final String contents )
    
        throws CoreException
        
    {
        writeFile( f, contents.getBytes( StandardCharsets.UTF_8 ) );
    }
    
    public static void writeFile( final File f,
                                  final byte[] contents )
    
        throws CoreException
        
    {
        writeFile( f, new ByteArrayInputStream( contents ) );
    }
    
    public static void writeFile( final File f,
                                  final InputStream contents )
    
        throws CoreException
        
    {
        if( f.exists() )
        {
            if( f.isDirectory() )
            {
                final String msg
                    = NLS.bind( Resources.locationIsDirectory, 
                                f.getAbsolutePath() );
                
                throw new CoreException( UtilPlugin.createErrorStatus( msg ) );
            }
        }
        else
        {
            mkdirs( f.getParentFile() );
        }
        
        final IFile wsfile = getWorkspaceFile( f );

        if( wsfile != null )
        {
            if( ! validateEdit( new IFile[] { wsfile } ).isOK() )
            {
                return;
            }
            
            if( wsfile.exists() )
            {
                wsfile.setContents( contents, true, false, null );
            }
            else
            {
                wsfile.create( contents, true, null );
            }
        }
        else
        {
            if( f.exists() && ! f.canWrite())
            {
                final String msg
                    = NLS.bind( Resources.cannotWriteFile, 
                                f.getAbsolutePath() );
                
                throw new CoreException( UtilPlugin.createErrorStatus( msg ) );
            }
            
            final byte[] buffer = new byte[ 1024 ];
            
            try( OutputStream out = new FileOutputStream( f ) )
            {
                for( int count; ( count = contents.read( buffer ) ) != -1; )
                {
                    out.write( buffer, 0, count );
                }
                
                out.flush();
            }
            catch( IOException e )
            {
                final String msg 
                    = NLS.bind( Resources.failedWhileWriting, f.getAbsolutePath() );
                
                throw new CoreException( UtilPlugin.createErrorStatus( msg ) );
            }
        }
    }
    
    public static final Comparator<IFile> getByNameIFileComparator()
    {
        final Comparator<IFile> comparator = new Comparator<IFile>()
        {
            public int compare( final IFile f1, 
                                final IFile f2 ) 
            {
                final String p1 = f1.getFullPath().toPortableString();
                final String p2 = f2.getFullPath().toPortableString();
                
                return p1.compareTo( p2 );
            }
        };
        
        return comparator;
    }
    
    /**
     * Verifies whether a resource does not exist, if an exact match is found returns.  Evaluates
     * whether the platform is caseSensitive and performs existence check accordingly.
     * @param f
     * @return
     */
    public static final boolean caseSensitiveDoesNotExist( final IFile f ) {
        if (f.exists()) {
            return false;
        }

        if( !caseSensitive ) {
            String fileName = f.getName();
            IContainer container = f.getParent();
            if (container.exists() && 
                    (container.getType() != IResource.PROJECT || ((IProject)container).isOpen())) {

                try {
                    for (IResource rsrc : container.members()) {
                        if (IResource.FILE == rsrc.getType()) {
                            if (fileName.equalsIgnoreCase(rsrc.getName())) {
                                return false;
                            }
                        }
                    }
                } catch (CoreException ce) {
                    return false;
                }
                
            }
        }
        return true;
    }
    
    /**
     * Determines the relative path between two resources
     * @param referencingResource
     * @param referencedResource
     * @return
     */
	public static String computeRelativeReferencePath(final IResource referencingResource, final IResource referencedResource) {
		if (referencingResource == null || referencedResource == null)
			return null;
		IPath referencingPath = referencingResource.getFullPath();
		IPath referencedPath = referencedResource.getFullPath();
		if (referencingPath.getDevice() != null || referencedPath.getDevice() != null)
			return null;
		referencingPath = referencingPath.makeRelative();
		referencedPath = referencedPath.makeRelative();

		// set up the minimum segment count needed to compute the relative path
		// for files, it should be 1 segment (for the filename)
		// for containers, it should be 0 segments
		int referencingSegmentCount = 1;
		int referencedSegmentCount = 1;
		if (referencingResource instanceof IContainer) {
			referencingSegmentCount = 0;
		}
		if (referencedResource instanceof IContainer) {
			referencedSegmentCount = 0;
		}
		while (referencingPath.segmentCount() > referencingSegmentCount && referencedPath.segmentCount() > referencedSegmentCount && referencingPath.segment(0).equals(referencedPath.segment(0))) {
			referencingPath = referencingPath.removeFirstSegments(1);
			referencedPath = referencedPath.removeFirstSegments(1);
		}

		IPath result = null;
		StringBuilder buf = new StringBuilder();
		String segment = null;
		do {
			segment = referencedPath.lastSegment();
			referencingPath = referencingPath.removeLastSegments(1);
			referencedPath = referencedPath.removeLastSegments(1);
			if (segment != null) {
				if (result == null)
					result = new Path(segment);
				else
					result = new Path(segment).append(result);
			}
			if (!referencingPath.equals(referencedPath) && !referencingPath.isEmpty())
				buf.append("../"); //$NON-NLS-1$
		} while (!referencingPath.equals(referencedPath));

		if (result != null)
			buf.append(result.makeRelative().toString());

		return buf.toString();
	}

	/**
	 * Computes the referenced file, given then referencing file and relative path.
	 * 
	 * This is the analog to the computeRelativeReferencePath() method, i.e., the
	 * return value of that method is the relative path passed to this method and
	 * the return value of this method is the referenced file passed to that method.
	 * 
	 * @param	referencingFile				Referencing file.
	 * @param	relativeReferencePath		Relative location/file from referencing file.
	 * @return								IFile referenced from file, following relative location.
	 */
	public static IFile computeReferencedFile(final IFile referencingFile, final String relativeReferencePath) {
		IPath wsdlPath = referencingFile.getParent().getFullPath().append(new Path(relativeReferencePath));
		return ResourcesPlugin.getWorkspace().getRoot().getFile(wsdlPath);
	}
	
    public static IStatus validateEdit( final IFile... files )
    {
        final IWorkspace ws = ResourcesPlugin.getWorkspace();
        return ws.validateEdit( files, IWorkspace.VALIDATE_PROMPT );
    }

    private static final class Resources
    
        extends NLS
        
    {
        public static String failedWhileWriting;
        public static String cannotWriteFile;
        public static String failedToCreateDirectory;
        public static String locationIsDirectory;
        public static String locationIsFile;
        
        static
        {
            initializeMessages( EclipseFileUtil.class.getName(), 
                                Resources.class );
        }
    }
    
    
}
