/******************************************************************************
 * Copyright (c) 2016 Oracle
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/

package oracle.eclipse.tools.util;

import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.sapphire.util.MutableReference;

/**
 * Contains utility functions to help deal with threading issues that are unique to Eclipse.
 * 
 * @author <a href="mailto:konstantin.komissarchik@oracle.com">Konstantin Komissarchik</a>
 */

public final class EclipseThreadingUtil
{
    /**
     * Prevents this static utility class from being instantiated.
     */
    
    private EclipseThreadingUtil() {}
    
    /**
     * Blocks the caller until general activity has settled down. This method returns when no
     * resources have been modified within the specified timeout period and when no jobs are
     * running or waiting to run.
     * 
     * @param timeout the timeout period to wait for resource changes (in seconds)
     */
    
    public static void waitForActivityToCease( final int timeout )
    {
        waitForResourceChangesToCease( timeout );
        waitForJobsActivityToCease();
    }
    
    /**
     * Blocks the caller until workspace resource change activity has settled down. The method
     * returns when no changes have been made within the specified timeout period.
     * 
     * @param timeout the timeout period to wait for changes (in seconds)
     */
    
    public static final void waitForResourceChangesToCease( final int timeout )
    {
        if( timeout < 0 )
        {
            throw new IllegalArgumentException();
        }
        
        final MutableReference<Integer> counter = new MutableReference<Integer>();
        counter.set( timeout );
        
        final IResourceChangeListener listener = new IResourceChangeListener()
        {
            public void resourceChanged( final IResourceChangeEvent event ) 
            {
                synchronized( counter )
                {
                    counter.set( timeout );
                }
            }
        };
        
        IWorkspace ws = ResourcesPlugin.getWorkspace();
        ws.addResourceChangeListener( listener, IResourceChangeEvent.POST_CHANGE );
        
        try
        {
            while( true )
            {
                synchronized( counter )
                {
                    if( counter.get() == 0 )
                    {
                        break;
                    }
                }
                
                try
                {
                    Thread.sleep( 1000 );
                    
                    synchronized( counter )
                    {
                        counter.set( counter.get() - 1 );
                    }
                    
                }
                catch( InterruptedException e )
                {
                    // Do not decrement the counter if the sleep was interrupted.
                }
            }
        }
        finally
        {
            // Remove the listener. Handle the case where the workspace has been
            // closed while this method was waiting.
            
            ws = ResourcesPlugin.getWorkspace();
            
            if( ws != null )
            {
                ws.removeResourceChangeListener( listener );
            }
        }
    }

    /**
     * Blocks the caller until there are no jobs running or waiting to run.
     */
    
    public static final void waitForJobsActivityToCease()
    {
        while( ! Job.getJobManager().isIdle() )
        {
            try
            {
                Thread.sleep( 1000 );
            }
            catch( InterruptedException e ) {}
        }
    }
    
}
