/******************************************************************************
 * Copyright (c) 2016 Oracle
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/

package oracle.eclipse.tools.util;

import static org.eclipse.jdt.core.IClasspathAttribute.JAVADOC_LOCATION_ATTRIBUTE_NAME;

import java.net.URL;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.jdt.apt.core.util.AptConfig;
import org.eclipse.jdt.apt.core.util.IFactoryPath;
import org.eclipse.jdt.core.IAccessRule;
import org.eclipse.jdt.core.IClasspathAttribute;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;

/**
 * Contains utility methods for querying and manipulating metadata associated
 * with a java project.
 */

public class JavaProjectUtil 

    extends ProjectUtil
    
{
    public static final String DOT_FACTORYPATH_FILE = ".factorypath"; //$NON-NLS-1$
    
    public static final String JRE_CONTAINER_ID 
        = "org.eclipse.jdt.launching.JRE_CONTAINER"; //$NON-NLS-1$
    
    private static final String JAR_PROTOCOL = "jar"; //$NON-NLS-1$
    private static final String JAR_EXTENSION = ".jar"; //$NON-NLS-1$
    private static final String ZIP_EXTENSION = ".zip"; //$NON-NLS-1$

    /**
     * Checks whether the specified project is a Java project.
     * 
     * @param pj the project to check
     * @return <code>true</code> if the project is a Java project
     */
    
    public static boolean isJavaProject( final IProject project )
    {
        try
        {
            return project.getNature( JavaCore.NATURE_ID ) != null;
        }
        catch( CoreException e ) {}

        return false;
    }
    
    /**
     * Checks whether the specified resource is inside a Java source folder of a Java project.
     * 
     * @param resource the resource to check
     * @return <code>true</code> if the resource is inside a Java source folder of a Java project
     */
    
    public static boolean isResourceInsideSourceFolder( final IResource resource )
    {
        final IProject proj = resource.getProject();
        
        if( isJavaProject( resource.getProject() ) )
        {
            final IPath path = resource.getFullPath();
            
            try
            {
                final IJavaProject jproj = JavaCore.create( proj );
                
                for( IClasspathEntry cpe : jproj.getRawClasspath() )
                {
                    if( cpe.getEntryKind() == IClasspathEntry.CPE_SOURCE && cpe.getPath().isPrefixOf( path ) )
                    {
                        return true;
                    }
                }
            }
            catch( CoreException e ) {}
        }
        
        return false;
    }
    
    /**
     * Replaces an entry on project classpath with another entry. Classpath order is preserved.
     * 
     * @param project the project
     * @param oldEntry the entry to find and replace
     * @param newEntry the entry to replace with
     * @throws CoreException if failed while retrieving or setting project classpath
     */
    
    public static void replaceClasspathEntry( final IProject project,
                                              final IClasspathEntry oldEntry,
                                              final IClasspathEntry newEntry )
    
        throws CoreException
        
    {
        if( isJavaProject( project ) )
        {
            replaceClasspathEntry( JavaCore.create( project ), oldEntry, newEntry );
        }
    }
    
    /**
     * Replaces an entry on project classpath with another entry. Classpath order is preserved.
     * 
     * @param project the project
     * @param oldEntry the entry to find and replace
     * @param newEntry the entry to replace with
     * @throws CoreException if failed while retrieving or setting project classpath
     */
    
    public static void replaceClasspathEntry( final IJavaProject project,
                                              final IClasspathEntry oldEntry,
                                              final IClasspathEntry newEntry )
    
        throws CoreException
        
    {
        final IClasspathEntry[] cp = project.getRawClasspath();
        
        for( int i = 0; i < cp.length; i++ )
        {
            if( cp[ i ].equals( oldEntry ) )
            {
                final IClasspathEntry[] newcp = new IClasspathEntry[ cp.length ];
                System.arraycopy( cp, 0, newcp, 0, cp.length );
                newcp[ i ] = newEntry;
                
                project.setRawClasspath( newcp, null );
                
                return;
            }
        }
    }
    
    public static void replaceAptFactoryPathPlugin( final IProject project,
                                                    final String oldPluginId,
                                                    final String newPluginId )

        throws CoreException

    {
        final IJavaProject jproj = JavaCore.create( project );
        
        if( jproj == null ) 
        {
            return;
        }
        
        final IFactoryPath ifp = AptConfig.getFactoryPath( jproj );
        ifp.disablePlugin( oldPluginId );
        ifp.enablePlugin( newPluginId );
        AptConfig.setFactoryPath( jproj, ifp );
        AptConfig.setEnabled( jproj, true );
    }
    
    public static IClasspathEntry newLibraryEntry( final IPath library )
    {
        return newLibraryEntry( library, null, null );
    }

    public static IClasspathEntry newLibraryEntry( final IPath library,
                                                   final IPath src,
                                                   final URL javadoc )
    {
        final IAccessRule[] access = {};
        
        final IClasspathAttribute[] attrs;
        
        if( javadoc == null )
        {
            attrs = new IClasspathAttribute[ 0 ];
        }
        else
        {
            String javadocUrlStr = javadoc.toExternalForm();
        
            if( ! javadoc.getProtocol().equals( JAR_PROTOCOL ) )
            {
                if( javadocUrlStr.endsWith( ZIP_EXTENSION ) ||
                    javadocUrlStr.endsWith( JAR_EXTENSION ) )
                {
                    javadocUrlStr = JAR_PROTOCOL + ":" + javadocUrlStr + "!/"; //$NON-NLS-1$ //$NON-NLS-2$
                }
            }
            
            attrs = new IClasspathAttribute[]
            { 
               JavaCore.newClasspathAttribute( JAVADOC_LOCATION_ATTRIBUTE_NAME, javadocUrlStr )
            };
        }
        
        return JavaCore.newLibraryEntry( library, src, null, access, attrs, false );
    }
    
    public static String getJavadocLocation( final IClasspathEntry cpe )
    {
        for( IClasspathAttribute cpattr : cpe.getExtraAttributes() )
        {
            if( cpattr.getName().equals( JAVADOC_LOCATION_ATTRIBUTE_NAME ) )
            {
                return cpattr.getValue();
            }
        }
        
        return null;
    }
    
}
