/******************************************************************************
 * Copyright (c) 2016 Oracle
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/

package oracle.eclipse.tools.util;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubProgressMonitor;

/**
 * @author <a href="mailto:konstantin.komissarchik@oracle.com">Konstantin Komissarchik</a>
 */

public final class ProgressMonitorUtil 
{
	private ProgressMonitorUtil()
	{
		// Not instantiable
	}
	
	/**
	 * Begins an unnamed task.
	 * 
	 * @param monitor a monitor or <code>null</code>
	 * @param totalWork the total work (use {@link IProgressMonitor#UNKNOWN} if unknown)
	 */
    public static void beginTask( final IProgressMonitor monitor,
                                  final String taskName,
                                  final int totalWork )
    {
        if( monitor != null )
        {
            monitor.beginTask( taskName, totalWork );
        }
    }
    
	/**
	 * Begins a named task.
	 * 
	 * @param monitor a monitor or <code>null</code>
	 * @param taskName the task name
	 * @param totalWork the total work (use {@link IProgressMonitor#UNKNOWN} if unknown)
	 */
    public static void beginTask( final IProgressMonitor monitor,
                                  final int totalWork )
    {
        beginTask( monitor, "", totalWork );
    }
    
	/**
	 * Reports a progress of a specified number of units of work
	 * 
	 * @param monitor a monitor or <code>null</code>
	 * @param work
	 */
    public static void worked( final IProgressMonitor monitor,
                               final int work )
    {
        if( monitor != null )
        {
            monitor.worked( work );
        }
    }
    
	/**
	 * Reports that the a task is done.
	 * 
	 * @param monitor a monitor or <code>null</code>
	 */
    public static void done( final IProgressMonitor monitor )
    {
        if( monitor != null )
        {
            monitor.done();
        }
    }
    
	/**
	 * Starts a sub-monitor for the specified parent monitor, specifying the number of '
	 * 'ticks'of the parent monitor are performed by the sub-monitor.
	 * 
	 * @param parent a monitor or <code>null</code>
	 * @param ticks
	 * @return a sub-monitor or <code>null</code> if parent is <code>null</code>
	 */
    public static IProgressMonitor submon( final IProgressMonitor parent,
                                            final int ticks )
    {
        return ( parent == null ? null : new SubProgressMonitor( parent, ticks ) );
    }
    
	/**
	 * Throws an {@link InterruptedException} is the monitor is {@link IProgressMonitor#isCanceled() cancelled}.
	 * 
	 * @param monitor a monitor or <code>null</code>
	 * @throws InterruptedException
	 */
    public static void checkIfCanceled( final IProgressMonitor monitor )
    
        throws InterruptedException
        
    {
        if( monitor != null && monitor.isCanceled() )
        {
            throw new InterruptedException();
        }
    }

	/**
	 * Sets whether the specified monitor is cancelled or not.
	 * 
	 * @param monitor a monitor or <code>null</code>
	 * @param cancel
	 */
	public static void setCanceled( IProgressMonitor monitor, boolean cancel ) 
	{
		if ( monitor != null ) 
		{
			monitor.setCanceled( cancel );
		}
	}
}
