/******************************************************************************
 * Copyright (c) 2016 Oracle
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/

package oracle.eclipse.tools.util;

import static oracle.eclipse.tools.util.ProgressMonitorUtil.submon;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.ICommand;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.osgi.util.NLS;

import oracle.eclipse.tools.util.internal.UtilPlugin;

/**
 * Contains utility methods for querying and manipulating metadata associated
 * with a basic project.
 */

public class ProjectUtil 
{
    public static final String DOT_PROJECT_FILE = ".project"; //$NON-NLS-1$
    
    public static class Node<T>
    {
        private T element = null;
        private List<Node<T>> children = new ArrayList<Node<T>>();
        
        private List<Node<T>> childrenReadOnly 
            = Collections.unmodifiableList( this.children );
        
        public synchronized T getElement()
        {
            return this.element;
        }
        
        public synchronized void setElement( final T element )
        {
            this.element = element;
        }
        
        public synchronized boolean hasChildren()
        {
            return ! this.children.isEmpty();
        }
        
        public synchronized Collection<Node<T>> getAllChildren()
        {
            final List<Node<T>> results = new ArrayList<Node<T>>();
            getAllChildren( results );
            return results;
        }
        
        private synchronized void getAllChildren( final List<Node<T>> results )
        {
            for( Node<T> node : this.children )
            {
                results.add( node );
                node.getAllChildren( results );
            }
        }
        
        public synchronized List<Node<T>> getChildren()
        {
            return this.childrenReadOnly;
        }
        
        public synchronized void setChildren( final List<Node<T>> children )
        {
            this.children.clear();
            this.children.addAll( children );
        }
        
        public synchronized Node<T> getChild( final int index )
        {
            return this.children.get( index );
        }
        
        public synchronized void setChild( final int index,
                                           final Node<T> node )
        {
            this.children.set( index, node );
        }
        
        public synchronized void addChild( final Node<T> child )
        {
            this.children.add( child );
        }
        
        public synchronized void removeChild( final Node<T> child )
        {
            this.children.remove( child );
        }
        
        public synchronized Node<T> convertDagToTree()
        {
            final Set<Node<T>> markedNodes = new HashSet<Node<T>>();
            return convertDagToTree( markedNodes );
        }
        
        private synchronized Node<T> convertDagToTree( final Set<Node<T>> markedNodes )
        {
            Node<T> node = this;
            
            if( markedNodes.contains( this ) )
            {
                node = new Node<T>();
                node.setElement( this.element );
                node.setChildren( this.children );
            }

            markedNodes.add( node );

            for( int i = 0, n = node.getChildren().size(); i < n; i++ )
            {
                final Node<T> x = node.getChild( i );
                final Node<T> y = x.convertDagToTree( markedNodes );
                
                if( x != y )
                {
                    node.setChild( i, y );
                }
            }
            
            return node;
        }
        
        public String toString()
        {
            final StringBuilder buf = new StringBuilder();
            toString( buf, "" );
            return buf.toString();
        }
        
        private void toString( final StringBuilder buf,
                               final String indent )
        {
            buf.append( indent );
            buf.append( ( this.element == null ? "<null>" : this.element.toString() ) );
            buf.append( ' ' );
            buf.append( hashCode() );
            buf.append( '\n' );
            
            final String childIndent = indent + "  ";
            
            for( Node<T> child : this.children )
            {
                child.toString( buf, childIndent );
            }
        }
    }
    
    public static final class CircularReferencesException
    
        extends Exception
        
    {
        private static final long serialVersionUID = 1L;
    }

    /**
     * Replaces one nature id with another in the project metadata. If the 
     * original nature id is not set on the project, this operation will
     * silently no-op.
     *  
     * @param project the project to perform nature replacement on
     * @param oldNatureId the original nature id to look for
     * @param newNatureId the new nature id to replace the original with
     * @throws CoreException if failed while replacing the nature
     */
    
    public static void replaceNature( final IProject project,
                                      final String oldNatureId,
                                      final String newNatureId )
    
        throws CoreException
        
    {
        final IProjectDescription pdesc = project.getDescription();
        
        final String[] oldNatureIds = pdesc.getNatureIds();
        String[] newNatureIds = null;
        
        for( int i = 0; i < oldNatureIds.length; i++ )
        {
            final String natureId = oldNatureIds[ i ];
            
            if( natureId.equals( oldNatureId ) )
            {
                if( newNatureIds == null )
                {
                    newNatureIds = new String[ oldNatureIds.length ];
                    System.arraycopy( oldNatureIds, 0, newNatureIds, 0, oldNatureIds.length );
                }
                
                newNatureIds[ i ] = newNatureId;
            }
        }
        
        if( newNatureIds != null )
        {
            pdesc.setNatureIds( newNatureIds );
            project.setDescription( pdesc, null );
        }
    }
    
	/**
	 * Adds nature to a project
	 */
	public static void addNature(IProject proj, String natureId) throws CoreException {
		IProjectDescription description = proj.getDescription();
		String[] prevNatures = description.getNatureIds();

		//check if the old nature does not exist
		boolean addNature = true;
		for (int i = 0; i < prevNatures.length; i++) {
			String nature = prevNatures[i];
			if (nature.equals(natureId)) {
				addNature = false;
				break;
			}
		}
		if (addNature) {
			String[] newNatures = new String[prevNatures.length + 1];
			System.arraycopy(prevNatures, 0, newNatures, 1, prevNatures.length);
			newNatures[0] = natureId;
			description.setNatureIds(newNatures);
			proj.setDescription(description, IResource.FORCE, null);
		}
	}
    
    /**
     * Remove one nature id from the project metadata. If the 
     * original nature id is not set on the project, this operation will
     * silently no-op.
     *  
     * @param project the project to perform nature removal on
     * @param oldNatureId the nature id to remove
     * @throws CoreException if failed while removing the nature
     */
    
    public static void removeNature( final IProject project,
                                      final String oldNatureId)
    
        throws CoreException
        
    {
        final IProjectDescription pdesc = project.getDescription();
        
        final String[] oldNatureIds = pdesc.getNatureIds();
        String[] newNatureIds = null;
        
        for( int i = 0; i < oldNatureIds.length; i++ )
        {
            final String natureId = oldNatureIds[ i ];
            
            if( natureId.equals( oldNatureId ) )
            {
                if( newNatureIds == null )
                {
                    newNatureIds = new String[ oldNatureIds.length - 1 ];
                    System.arraycopy( oldNatureIds, 0, newNatureIds, 0, i );
                    if (i < oldNatureIds.length - 1)
                    {
                    	System.arraycopy( oldNatureIds, i + 1, newNatureIds, i, oldNatureIds.length - i - 1);
                    }
                    break;
                }                
            }
        }
        
        if( newNatureIds != null )
        {
            pdesc.setNatureIds( newNatureIds );
            project.setDescription( pdesc, null );
        }
    }
    
    public static boolean hasBuilder( final IProject project,
                                      final String builderId )
    
        throws CoreException
        
    {
        final IProjectDescription pdesc = project.getDescription();
        final ICommand[] existingBuildSpec = pdesc.getBuildSpec();
        
        for( ICommand cmd : existingBuildSpec )
        {
            final String bid = cmd.getBuilderName();
            
            if( bid != null && bid.equals( builderId ) )
            {
                return true;
            }
        }
        
        return false;
    }
    
    public static void addBuilder( final IProject project,
                                   final String builderId )
    
        throws CoreException
        
    {
        final IProjectDescription pdesc = project.getDescription();
        final ICommand[] existingBuildSpec = pdesc.getBuildSpec();
        
        for( ICommand cmd : existingBuildSpec )
        {
            if( cmd.getBuilderName().equals( builderId ) )
            {
                return;
            }
        }
        
        final ICommand[] newBuildSpec = new ICommand[ existingBuildSpec.length + 1 ];
        System.arraycopy( existingBuildSpec, 0, newBuildSpec, 0, existingBuildSpec.length );
        final ICommand newCommand = pdesc.newCommand();
        newCommand.setBuilderName( builderId );
        newBuildSpec[ newBuildSpec.length - 1 ] = newCommand;
        pdesc.setBuildSpec( newBuildSpec );
        
        project.setDescription( pdesc, null );
    }

    public static void removeBuilder( final IProject project,
                                      final String builderId )
    
        throws CoreException
        
    {
        final IProjectDescription pdesc = project.getDescription();
        final List<ICommand> newBuildSpec = new ArrayList<ICommand>();
        
        for( ICommand cmd : pdesc.getBuildSpec() )
        {
            final String bid = cmd.getBuilderName();
            
            if( bid == null || ! bid.equals( builderId ) )
            {
                newBuildSpec.add( cmd );
            }
        }
        
        final ICommand[] newBuildSpecArray 
            = newBuildSpec.toArray( new ICommand[ newBuildSpec.size() ] ); 

        pdesc.setBuildSpec( newBuildSpecArray );
        project.setDescription( pdesc, null );
    }

    /**
     * Run the specified builder on a project.
     * 
     * @param project the project
     * @param builderId the builder id
     * @param monitor progress monitor 
     */
    
    public static void runBuilder( final IProject project, 
                                   final String builderId, 
                                   final IProgressMonitor monitor )
    
        throws CoreException 
    
    {
        final ICommand builder = getCommand(project, builderId);
        if (builder != null) {
            project.build(IncrementalProjectBuilder.INCREMENTAL_BUILD, builderId, builder
                    .getArguments(), monitor);
        }
    }
    
    public static void scheduleIncrementalBuild( final Set<IProject> projects )
    {
        scheduleBuild( projects, IncrementalProjectBuilder.INCREMENTAL_BUILD );
    }
    
    public static void scheduleFullBuild( final Set<IProject> projects )
    {
        scheduleBuild( projects, IncrementalProjectBuilder.FULL_BUILD );
    }
    
    public static void scheduleCleanBuild( final Set<IProject> projects )
    {
        scheduleBuild( projects, IncrementalProjectBuilder.CLEAN_BUILD );
    }
    
    public static void scheduleBuild( final Set<IProject> projects,
                                      final int buildKind )
    {
        final IWorkspace ws = ResourcesPlugin.getWorkspace();
        
        final Job job = new WorkspaceJob( Resources.buildJobName )
        {
            @Override
            public boolean belongsTo( final Object family ) 
            {
                return family.equals( ResourcesPlugin.FAMILY_MANUAL_BUILD );
            }
            
            @Override
            public IStatus runInWorkspace( final IProgressMonitor monitor )
            
                throws CoreException
                
            {
                MultiStatus errors = null;
                
                monitor.beginTask( "", projects.size() * 1000 );
                
                try
                {
                    final IProject[] pjarray = projects.toArray( new IProject[ projects.size() ] );
                    final IWorkspace.ProjectOrder order = ws.computeProjectOrder( pjarray );
                    
                    for( IProject project : order.projects )
                    {
                        try
                        {
                            project.build( buildKind, submon( monitor, 1000 ) );
                        }
                        catch( CoreException e )
                        {
                            errors = recordError( errors, e );
                        }
                    }
                }
                finally
                {
                    monitor.done();
                }
                
                return errors == null ? Status.OK_STATUS : errors;
            }
            
            private MultiStatus recordError( MultiStatus errors, 
                                             final CoreException error ) 
            {
                if( errors == null ) 
                {
                    errors = new MultiStatus( UtilPlugin.SYMBOLIC_NAME, IStatus.ERROR, Resources.buildFailed, null );
                }
                
                errors.merge( error.getStatus() );
                
                return errors;
            }
        };
        
        job.setRule( ws.getRuleFactory().buildRule() );
        job.schedule();
    }

    /**
     * Get a specific command in the build spec of a project.
     * 
     * @param project
     *            a project
     * @param builderId
     *            a builder id
     * @return the corresponding command from the project build spec, or
     *         <code>null</code> if none
     */
    
    private static ICommand getCommand(IProject project, String builderId) throws CoreException {
        IProjectDescription description = project.getDescription();
        if (description == null) {
            return null;
        }

        ICommand[] commands = description.getBuildSpec();
        for (int i = 0; i < commands.length; ++i) {
            if (commands[i].getBuilderName().equals(builderId)) {
                return commands[i];
            }
        }
        return null;
    }
    
    public static List<Node<IProject>> getProjectReferenceGraph()
    
        throws CoreException, CircularReferencesException
        
    {
        // Build information about projects and what they reference.
        
        final Map<IProject,Set<IProject>> projectReferences
            = new HashMap<IProject,Set<IProject>>();
        
        final IWorkspaceRoot wsroot = ResourcesPlugin.getWorkspace().getRoot();
        
        for( IProject project : wsroot.getProjects() )
        {
            if( ! project.isAccessible() )
            {
                continue;
            }
            
            final Set<IProject> references = new HashSet<IProject>();
            projectReferences.put( project, references );
            
            for( IProject refpj : project.getDescription().getReferencedProjects() )
            {
                if( refpj.isAccessible() )
                {
                    references.add( refpj );
                }
            }
            
            if( JavaProjectUtil.isJavaProject( project ) )
            {
                final IJavaProject jproj = JavaCore.create( project );

                for( IClasspathEntry cpe : jproj.getResolvedClasspath( true ) )
                {
                    if( cpe.getEntryKind() == IClasspathEntry.CPE_PROJECT )
                    {
                        final String refpjname = cpe.getPath().segment( 0 );
                        final IProject refpj = wsroot.getProject( refpjname );
                        
                        if( refpj.isAccessible() )
                        {
                            references.add( refpj );
                        }
                    }
                }
            }
        }
        
        // Construct the reference graph from bottom up.
        
        final List<Node<IProject>> graph = new ArrayList<Node<IProject>>();
        
        while( ! projectReferences.isEmpty() )
        {
            final List<IProject> toRemove = new ArrayList<IProject>();
            
            for( Map.Entry<IProject,Set<IProject>> entry : projectReferences.entrySet() )
            {
                final IProject project = entry.getKey();
                final List<Node<IProject>> refnodes = new ArrayList<Node<IProject>>();
                
                boolean foundAll = true;
                
                for( IProject refpj : entry.getValue() )
                {
                    final Node<IProject> refpjnode = findProjectInGraph( graph, refpj );
                    
                    if( refpjnode == null )
                    {
                        foundAll = false;
                        break;
                    }
                    
                    refnodes.add( refpjnode );
                }
                
                if( foundAll )
                {
                    final Node<IProject> pjnode = new Node<IProject>();
                    pjnode.setElement( project );
                    pjnode.setChildren( refnodes );
                    
                    graph.removeAll( refnodes );
                    graph.add( pjnode );
                    toRemove.add( project );
                }
            }
            
            if( toRemove.isEmpty() )
            {
                throw new CircularReferencesException();
            }
            
            for( IProject project : toRemove )
            {
                projectReferences.remove( project );
            }
        }
        
        return graph;
    }
    
    private static Node<IProject> findProjectInGraph( final List<Node<IProject>> graph,
                                                      final IProject project )
    {
        for( Node<IProject> node : graph )
        {
            if( node.getElement().equals( project ) )
            {
                return node;
            }
            
            final Node<IProject> res = findProjectInGraph( node.getChildren(), project );
            
            if( res != null )
            {
                return res;
            }
        }
        
        return null;
    }
    
    private static final class Resources
        
        extends NLS

    {
        public static String buildJobName;
        public static String buildFailed;
        
        static
        {
            initializeMessages( ProjectUtil.class.getName(), Resources.class );
        }
    }

}
