/******************************************************************************
 * Copyright (c) 2016 Oracle
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/

package oracle.eclipse.tools.util;

import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.eclipse.core.runtime.preferences.IEclipsePreferences;
import org.eclipse.core.runtime.preferences.InstanceScope;
import org.eclipse.sapphire.util.ListFactory;
import org.osgi.service.prefs.BackingStoreException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Interface for directly reading and writing WTP runtime definitions. It is intended
 * for cases not covered by WTP's public API, such as updating runtime definitions
 * when runtime types are changed across releases and the old types are no longer
 * available in the current Eclipse instance.
 * 
 * @author <a href="mailto:konstantin.komissarchik@oracle.com">Konstantin Komissarchik</a>
 */

public final class WtpRuntimesMetadata 
{
	private static final String SERVER_CORE_PREF = "org.eclipse.wst.server.core";
	private static final String SERVER_CORE_PREF_RUNTIMES = "runtimes";
	
    private static final String EL_RUNTIME = "runtime";
    
    private static final String ATTR_ID = "id";
    private static final String ATTR_NAME = "name";
    private static final String ATTR_RUNTIME_TYPE_ID = "runtime-type-id";

    private final IEclipsePreferences preference;
    private final Document document;

    public WtpRuntimesMetadata() throws IOException
    {
		this.preference = InstanceScope.INSTANCE.getNode( SERVER_CORE_PREF );
		final String runtimeString = this.preference.get( SERVER_CORE_PREF_RUNTIMES, null );
		
		if( runtimeString == null )
		{
		    this.document = null;
		}
		else
		{
            try
            {
                final DocumentBuilder docbuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
                this.document = docbuilder.parse( new InputSource( new StringReader( runtimeString ) ) );
            }
            catch( final ParserConfigurationException | SAXException e )
            {
                throw new RuntimeException( e );
            }
		}
    }
    
    public List<Runtime> runtimes()
    {
        final ListFactory<Runtime> runtimes = ListFactory.start();
        
        if( this.document != null )
        {
            final Element root = this.document.getDocumentElement();
            
            if( root != null )
            {
                final NodeList children = root.getChildNodes();
                
                for( int i = 0, n = children.getLength(); i < n; i++ )
                {
                    final Node child = children.item( i );
                    
                    if( child instanceof Element && child.getNodeName().equals( EL_RUNTIME ) )
                    {
                        runtimes.add( new Runtime( (Element) child ) );
                    }
                }
            }
        }
        
        return runtimes.result();
    }
    
    public void save() throws IOException
    {
        try
        {
            final StringWriter sw = new StringWriter();
            
            final DOMSource source = new DOMSource( this.document );
            final StreamResult result = new StreamResult( sw );
            
            final TransformerFactory factory = TransformerFactory.newInstance();
            final Transformer transformer = factory.newTransformer();
            
            transformer.transform( source, result );
            
            this.preference.put( SERVER_CORE_PREF_RUNTIMES, sw.toString() );
            this.preference.flush();
        }
        catch( final BackingStoreException e )
        {
            throw new IOException( e );
        }
        catch( final TransformerException e )
        {
            throw new RuntimeException( e );
        }
    }
    
    public static final class Runtime
    {
        private final Element element;
        
        public Runtime( final Element element )
        {
            this.element = element;
        }
        
        public String getId()
        {
            return this.element.getAttribute( ATTR_ID );
        }
        
        public void setId( final String id )
        {
            this.element.setAttribute( ATTR_ID, id );
        }
        
        public String getName()
        {
            return this.element.getAttribute( ATTR_NAME );
        }
        
        public void setName( final String name )
        {
            this.element.setAttribute( ATTR_NAME, name );
        }
        
        public String getRuntimeType()
        {
            return this.element.getAttribute( ATTR_RUNTIME_TYPE_ID );
        }
        
        public void setRuntimeType( final String runtimeType )
        {
            this.element.setAttribute( ATTR_RUNTIME_TYPE_ID, runtimeType );
        }
    }
    
}
