/******************************************************************************
 * Copyright (c) 2016 Oracle
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 ******************************************************************************/

package oracle.eclipse.tools.util;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.sapphire.util.ListFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Interface for directly reading and writing WTP server definitions. It is intended
 * for cases not covered by WTP's public API, such as updating server definitions
 * when server types are changed across releases and the old types are no longer
 * available in the current Eclipse instance.
 * 
 * @author <a href="mailto:konstantin.komissarchik@oracle.com">Konstantin Komissarchik</a>
 */

public final class WtpServersMetadata 
{
    private static final String EL_SERVER = "server";
    
    private static final String ATTR_ID = "id";
    private static final String ATTR_NAME = "name";
    private static final String ATTR_SERVER_TYPE = "server-type";
    private static final String ATTR_SERVER_TYPE_ID = "server-type-id";

    private final File file;
    private final Document document;

    public WtpServersMetadata() throws IOException
    {
        this.file = ResourcesPlugin.getWorkspace().getRoot().getLocation().append( ".metadata/.plugins/org.eclipse.wst.server.core/servers.xml" ).toFile();
        
        if( this.file.exists() )
        {
            try( Reader reader = new FileReader( this.file ) )
            {
                final DocumentBuilder docbuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
                this.document = docbuilder.parse( new InputSource( reader ) );
            }
            catch( final ParserConfigurationException | SAXException e )
            {
                throw new RuntimeException( e );
            }
        }
        else
        {
            this.document = null;
        }
    }
    
    public List<Server> servers()
    {
        final ListFactory<Server> servers = ListFactory.start();
        
        if( this.document != null )
        {
            final Element root = this.document.getDocumentElement();
            
            if( root != null )
            {
                final NodeList children = root.getChildNodes();
                
                for( int i = 0, n = children.getLength(); i < n; i++ )
                {
                    final Node child = children.item( i );
                    
                    if( child instanceof Element && child.getNodeName().equals( EL_SERVER ) )
                    {
                        servers.add( new Server( (Element) child ) );
                    }
                }
            }
        }
        
        return servers.result();
    }
    
    public void save() throws IOException
    {
        try( FileWriter writer = new FileWriter( this.file ) )
        {
            final DOMSource source = new DOMSource( this.document );
            final StreamResult result = new StreamResult( writer );
            
            final TransformerFactory factory = TransformerFactory.newInstance();
            final Transformer transformer = factory.newTransformer();
            
            transformer.transform( source, result );
        }
        catch( final TransformerException e )
        {
            throw new RuntimeException( e );
        }
    }
    
    public static final class Server
    {
        private final Element element;
        
        public Server( final Element element )
        {
            this.element = element;
        }
        
        public String getId()
        {
            return this.element.getAttribute( ATTR_ID );
        }
        
        public void setId( final String id )
        {
            this.element.setAttribute( ATTR_ID, id );
        }
        
        public String getName()
        {
            return this.element.getAttribute( ATTR_NAME );
        }
        
        public void setName( final String name )
        {
            this.element.setAttribute( ATTR_NAME, name );
        }
        
        public String getServerType()
        {
            return this.element.getAttribute( ATTR_SERVER_TYPE );
        }
        
        public void setServerType( final String serverType )
        {
            this.element.setAttribute( ATTR_SERVER_TYPE, serverType );
            this.element.setAttribute( ATTR_SERVER_TYPE_ID, serverType );
        }
    }
    
}
